"""Mission generator."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import ChainMap, OrderedDict
from dataclasses import fields
from typing import Dict, List, Mapping, Optional, Union

import pandas as pd

from fastoad.constants import EngineSetting
from fastoad.model_base import FlightPoint
from fastoad.model_base.propulsion import IPropulsion
from fastoad.openmdao.variables import VariableList
from .constants import NAME_TAG, SEGMENT_TYPE_TAG, TYPE_TAG
from .input_definition import InputDefinition
from .structure_builders import (
    AbstractStructureBuilder,
    MissionStructureBuilder,
    PhaseStructureBuilder,
)
from ..exceptions import FastMissionFileMissingMissionNameError
from ..schema import (
    CLIMB_PARTS_TAG,
    CRUISE_PART_TAG,
    DESCENT_PARTS_TAG,
    MISSION_DEFINITION_TAG,
    MissionDefinition,
    PARTS_TAG,
    PHASE_TAG,
    RESERVE_TAG,
    ROUTE_TAG,
    SEGMENT_TAG,
)
from ...base import FlightSequence
from ...polar import Polar
from ...routes import RangedRoute
from ...segments.base import AbstractFlightSegment, SegmentDefinitions


class MissionBuilder:
    """
    This class builds and computes a mission from a provided definition.
    """

    def __init__(
        self,
        mission_definition: Union[str, MissionDefinition],
        *,
        propulsion: IPropulsion = None,
        reference_area: float = None,
    ):
        """
        :param mission_definition: a file path or MissionDefinition instance
        :param propulsion: if not provided, the property :attr:`propulsion` must be
                           set before calling :meth:`build`
        :param reference_area: if not provided, the property :attr:`reference_area` must be
                               set before calling :meth:`build`
        """
        self._structure_builders: Dict[str, AbstractStructureBuilder] = {}
        self.definition = mission_definition
        self._base_kwargs = {"reference_area": reference_area, "propulsion": propulsion}

    @property
    def definition(self) -> MissionDefinition:
        """
        The mission definition instance.

        If it is set as a file path, then the matching file will be read and interpreted.
        """
        return self._definition

    @definition.setter
    def definition(self, mission_definition: Union[str, MissionDefinition]):
        if isinstance(mission_definition, str):
            self._definition = MissionDefinition(mission_definition)
        else:
            self._definition = mission_definition

        for mission_name in self._definition[MISSION_DEFINITION_TAG]:
            self._structure_builders[mission_name] = MissionStructureBuilder(
                self._definition, mission_name
            )

            if self.get_input_weight_variable_name(mission_name) is None:
                self._add_default_taxi_takeoff(mission_name)

    @property
    def propulsion(self) -> IPropulsion:
        """Propulsion model for performance computation."""
        return self._base_kwargs["propulsion"]

    @propulsion.setter
    def propulsion(self, propulsion: IPropulsion):
        self._base_kwargs["propulsion"] = propulsion

    @property
    def reference_area(self) -> float:
        """Reference area for aerodynamic polar."""
        return self._base_kwargs["reference_area"]

    @reference_area.setter
    def reference_area(self, reference_area: float):
        self._base_kwargs["reference_area"] = reference_area

    def build(self, inputs: Optional[Mapping] = None, mission_name: str = None) -> FlightSequence:
        """
        Builds the flight sequence from definition file.

        :param inputs: if provided, any input parameter that is a string which matches
                       a key of `inputs` will be replaced by the corresponding value
        :param mission_name: mission name (can be omitted if only one mission is defined)
        :return:
        """
        if self.get_input_weight_variable_name(mission_name) is None:
            self._add_default_taxi_takeoff(mission_name)

        for input_def in self._structure_builders[mission_name].get_input_definitions():
            input_def.set_variable_value(inputs)
        if mission_name is None:
            mission_name = self.get_unique_mission_name()
        mission = self._build_mission(self._structure_builders[mission_name].structure)
        return mission

    def get_route_ranges(
        self, inputs: Optional[Mapping] = None, mission_name: str = None
    ) -> List[float]:
        """

        :param inputs: if provided, any input parameter that is a string which matches
                       a key of `inputs` will be replaced by the corresponding value
        :param mission_name: mission name (can be omitted if only one mission is defined)
        :return: list of flight ranges for each element of the flight sequence that is a route
        """
        routes = self.build(inputs, mission_name).flight_sequence
        return [route.flight_distance for route in routes if isinstance(route, RangedRoute)]

    def get_reserve(self, flight_points: pd.DataFrame, mission_name: str = None) -> float:
        """
        Computes the reserve fuel according to definition in mission input file.

        :param flight_points: the dataframe returned by compute_from() method of the
                              instance returned by :meth:`build`
        :param mission_name: mission name (can be omitted if only one mission is defined)
        :return: the reserve fuel mass in kg, or 0.0 if no reserve is defined.
        """

        if mission_name is None:
            mission_name = self.get_unique_mission_name()

        last_part_spec = self._get_mission_part_structures(mission_name)[-1]
        if RESERVE_TAG in last_part_spec:
            ref_name = last_part_spec[RESERVE_TAG]["ref"].value
            multiplier = last_part_spec[RESERVE_TAG]["multiplier"].value

            route_points = flight_points.loc[
                flight_points.name.str.contains(f"{mission_name}:{ref_name}")
            ]
            consumed_mass = route_points.mass.iloc[0] - route_points.mass.iloc[-1]
            return consumed_mass * multiplier

        return 0.0

    def get_input_variables(self, mission_name=None) -> VariableList:
        """
        Identify variables for a defined mission.

        :param mission_name: mission name (can be omitted if only one mission is defined)
        :return: a VariableList instance.
        """
        if mission_name is None:
            mission_name = self.get_unique_mission_name()

        input_definition = VariableList()
        for input_def in self._structure_builders[mission_name].get_input_definitions():
            if input_def.variable_name and input_def.variable_name not in input_definition.names():
                input_definition.append(input_def.get_input_definition())

        return input_definition

    def get_unique_mission_name(self) -> str:
        """
        Provides mission name if only one mission is defined in mission file.

        :return: the mission name, if only one mission is defined
        :raise FastMissionFileMissingMissionNameError: if several missions are defined in mission
                                                       file
        """
        if len(self._structure_builders) == 1:
            return list(self._structure_builders.keys())[0]

        raise FastMissionFileMissingMissionNameError(
            "Mission name must be specified if several missions are defined in mission file."
        )

    def get_input_weight_variable_name(self, mission_name: str) -> Optional[str]:
        """
        Search the mission structure for a segment that has a target absolute mass defined and
        returns the associated variable name.

        :param mission_name:
        :return: The variable name, or None if no target mass found.
        """
        return self._get_input_weight_variable_name_in_structure(
            self._get_mission_part_structures(mission_name)
        )

    def get_mission_start_mass_input(self, mission_name: str) -> Optional[str]:
        """

        :param mission_name:
        :return: Target mass variable of first segment, if any.
        """
        part = self._get_first_segment_structure(mission_name)
        if "mass" in part["target"]:
            return part["target"]["mass"].variable_name

        return None

    def _get_first_segment_structure(self, mission_name: str):
        part = self._get_mission_part_structures(mission_name)[0]
        while PARTS_TAG in part:
            part = part[PARTS_TAG][0]
        return part

    def get_mission_part_names(self, mission_name: str) -> List[str]:
        """

        :param mission_name:
        :return: list of names of parts (phase or route) for specified mission.
        """
        return [
            part[NAME_TAG]
            for part in self._get_mission_part_structures(mission_name)
            if part.get(TYPE_TAG) in [ROUTE_TAG, PHASE_TAG]
        ]

    def _build_mission(self, mission_structure: OrderedDict) -> FlightSequence:
        """
        Builds mission instance from provided structure.

        :param mission_structure: structure of the mission to build
        :return: the mission instance
        """
        mission = FlightSequence()

        part_kwargs = self._get_part_kwargs({}, mission_structure)

        mission.name = mission_structure[NAME_TAG]
        for part_spec in mission_structure[PARTS_TAG]:
            if TYPE_TAG not in part_spec:
                continue
            if part_spec[TYPE_TAG] == SEGMENT_TAG:
                part = self._build_segment(part_spec, part_kwargs)
            if part_spec[TYPE_TAG] == ROUTE_TAG:
                part = self._build_route(part_spec, part_kwargs)
            elif part_spec[TYPE_TAG] == PHASE_TAG:
                part = self._build_phase(part_spec, part_kwargs)
            mission.flight_sequence.append(part)

        return mission

    def _build_route(self, route_structure: OrderedDict, kwargs: Mapping = None):
        """
        Builds route instance.

        :param route_structure: structure of the route to build
        :return: the route instance
        """
        climb_phases = []
        descent_phases = []

        part_kwargs = self._get_part_kwargs(
            kwargs,
            route_structure,
            [
                CLIMB_PARTS_TAG,
                CRUISE_PART_TAG,
                DESCENT_PARTS_TAG,
                "range",
                "distance_accuracy",
            ],
        )

        for part_structure in route_structure[CLIMB_PARTS_TAG]:
            phase = self._build_phase(part_structure, part_kwargs)
            climb_phases.append(phase)

        cruise_phase = self._build_segment(
            route_structure[CRUISE_PART_TAG],
            ChainMap({"target": FlightPoint(ground_distance=0.0)}, part_kwargs),
        )

        for part_structure in route_structure[DESCENT_PARTS_TAG]:
            phase = self._build_phase(part_structure, part_kwargs)
            descent_phases.append(phase)

        if "range" in route_structure:
            flight_range = route_structure["range"].value
            route = RangedRoute(
                climb_phases=climb_phases,
                cruise_segment=cruise_phase,
                descent_phases=descent_phases,
                flight_distance=flight_range,
            )
        else:
            route = FlightSequence()
            route.flight_sequence.extend(climb_phases)

        if "distance_accuracy" in route_structure:
            route.distance_accuracy = route_structure["distance_accuracy"].value

        route.name = route_structure[NAME_TAG]
        return route

    def _build_phase(self, phase_structure: Mapping, kwargs: Mapping = None):
        """
        Builds phase instance

        :param phase_structure: structure of the phase to build
        :return: the phase instance
        """
        phase = FlightSequence(name=phase_structure[NAME_TAG])
        part_kwargs = self._get_part_kwargs(kwargs, phase_structure)

        for part_structure in phase_structure[PARTS_TAG]:
            if part_structure[TYPE_TAG] == PHASE_TAG:
                flight_part = self._build_phase(part_structure, part_kwargs)
            else:
                flight_part = self._build_segment(part_structure, part_kwargs)
            phase.flight_sequence.append(flight_part)

        return phase

    def _build_segment(self, segment_definition: Mapping, kwargs: Mapping) -> AbstractFlightSegment:
        """
        Builds a flight segment according to provided definition.

        :param segment_definition: the segment definition from mission file
        :param kwargs: a preset of keyword arguments for AbstractFlightSegment instantiation
        :param tag: the expected tag for specifying the segment type
        :return: the FlightSegment instance
        """
        segment_class = SegmentDefinitions.get_segment_class(segment_definition[SEGMENT_TYPE_TAG])
        part_kwargs = kwargs.copy()
        part_kwargs.update(segment_definition)
        part_kwargs.update(self._base_kwargs)
        for key, value in part_kwargs.items():
            if key == "polar":
                value = Polar(value["CL"].value, value["CD"].value)
            elif key == "target":
                if not isinstance(value, FlightPoint):
                    target_parameters = {
                        param.parameter_name: param.value for param in value.values()
                    }
                    relative_fields = [
                        param.parameter_name for param in value.values() if param.is_relative
                    ]
                    value = FlightPoint(**target_parameters)
                    value.set_as_relative(relative_fields)

            part_kwargs[key] = value

        self._replace_input_definitions_by_values(part_kwargs)

        if "engine_setting" in part_kwargs:
            part_kwargs["engine_setting"] = EngineSetting.convert(part_kwargs["engine_setting"])

        input_field_names = [
            class_field.name for class_field in fields(segment_class) if class_field.init
        ]
        part_kwargs = {key: value for key, value in part_kwargs.items() if key in input_field_names}
        segment = segment_class(**part_kwargs)
        return segment

    @staticmethod
    def _replace_input_definitions_by_values(part_kwargs):
        for key, input_def in part_kwargs.items():
            if isinstance(input_def, InputDefinition):
                part_kwargs[key] = input_def.value

    def _get_mission_part_structures(self, mission_name: str):
        return self._structure_builders[mission_name].structure[PARTS_TAG]

    def _get_part_kwargs(
        self, kwargs: Mapping, phase_structure: Mapping, specific_exclude: List[str] = None
    ):
        part_kwargs = {}
        if kwargs is not None:
            part_kwargs.update(kwargs)
        if specific_exclude is None:
            specific_exclude = []
        exclude = [NAME_TAG, PARTS_TAG, TYPE_TAG] + specific_exclude
        part_kwargs.update(
            {name: value for name, value in phase_structure.items() if name not in exclude}
        )
        self._replace_input_definitions_by_values(part_kwargs)
        return part_kwargs

    def _get_input_weight_variable_name_in_structure(
        self, structure: Union[list, dict]
    ) -> Optional[str]:
        for tag in [PARTS_TAG, CLIMB_PARTS_TAG, DESCENT_PARTS_TAG]:
            if tag in structure:
                return self._get_input_weight_variable_name_in_structure(structure[tag])

        if isinstance(structure, list):
            for item in structure:
                result = self._get_input_weight_variable_name_in_structure(item)
                if result:
                    return result

        if "target" in structure and "mass" in structure["target"]:
            target_mass = structure["target"]["mass"]
            if not target_mass.is_relative:
                name = target_mass.variable_name
                if name:
                    return name

        return None

    def _add_default_taxi_takeoff(self, mission_name):
        definition = {
            "phases": {
                "taxi_out": {
                    "parts": [
                        {
                            "segment": "start",
                            "target": {
                                "altitude": {"value": "~", "unit": "ft", "default": 0.0},
                                "true_airspeed": {"value": "~", "unit": "m/s", "default": 0.0},
                            },
                        },
                        {
                            "segment": "taxi",
                            "time_step": {
                                "value": "settings:mission~",
                                "default": 60.0,
                                "unit": "s",
                            },
                            "thrust_rate": "~",
                            "true_airspeed": {"value": "~", "unit": "m/s", "default": 0.0},
                            "target": {"time": "~duration"},
                        },
                    ],
                },
                "takeoff": {
                    "parts": [
                        {
                            "segment": "transition",
                            "target": {
                                "delta_altitude": {
                                    "value": "~safety_altitude",
                                    "unit": "ft",
                                    "default": 35.0,
                                },
                                "delta_mass": {"value": "-~fuel", "unit": "kg"},
                                "time": {"value": "~duration", "default": 0.0, "unit": "s"},
                                "true_airspeed": {"value": "~V2", "unit": "m/s"},
                            },
                        },
                        {
                            "segment": "mass_input",
                            "target": {"mass": {"value": "~~:TOW", "unit": "kg"}},
                        },
                    ]
                },
            }
        }

        taxi_out = PhaseStructureBuilder(definition, "taxi_out", mission_name)
        taxi_out_structure = self._structure_builders[mission_name].process_builder(taxi_out)
        self._structure_builders[mission_name].structure[PARTS_TAG].insert(0, taxi_out_structure)

        takeoff = PhaseStructureBuilder(definition, "takeoff", mission_name)
        takeoff_structure = self._structure_builders[mission_name].process_builder(takeoff)
        self._structure_builders[mission_name].structure[PARTS_TAG].insert(1, takeoff_structure)



"""
Class for managing a list of OpenMDAO variables.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import itertools
from copy import deepcopy
from typing import Iterable, List, Mapping, Tuple, Union

import numpy as np
import openmdao.api as om
import pandas as pd
from deprecated import deprecated
from openmdao.core.constants import _SetupStatus

from fastoad.openmdao._utils import get_unconnected_input_names, problem_without_mpi
from .variable import METADATA_TO_IGNORE, Variable


class VariableList(list):
    """
    Class for storing OpenMDAO variables.

    A list of :class:`~fastoad.openmdao.variables.variable.Variable` instances, but items can
    also be accessed through variable names. It also has utilities to be converted from/to some
    other data structures (python dict, OpenMDAO IndepVarComp, pandas DataFrame)

    See documentation of :class:`~fastoad.openmdao.variables.variable.Variable` to see how to
    manipulate each element.

    There are several ways for adding variables::

        # Assuming these Python variables are ready...
        var_1 = Variable('var/1', value=0.)
        metadata_2 = {'value': 1., 'units': 'm'}

        # ... a VariableList instance can be populated like this:
        vars_A = VariableList()
        vars_A.append(var_1)              # Adds directly a Variable instance
        vars_A['var/2'] = metadata_2      # Adds the variable with given name and given metadata

    Note:
        Adding a Variable instance with a name that is already in the VariableList instance
        will replace the previous Variable instance instead of adding a new one.

    .. code:: python

        # It is also possible to instantiate a VariableList instance from another VariableList
        # instance or a simple list of Variable instances
        vars_B = VariableList(vars_A)
        vars_C = VariableList([var_1])

        # An existing VariableList instance can also receive the content of another VariableList
        # instance.
        vars_C.update(vars_A)             # variables in vars_A will overwrite variables with same
                                          # name in vars_C

    After that, following equalities are True::

        print( var_1 in vars_A )
        print( 'var/1' in vars_A.names() )
        print( 'var/2' in vars_A.names() )
    """

    def names(self) -> List[str]:
        """
        :return: names of variables
        """
        return [var.name for var in self]

    def metadata_keys(self) -> List[str]:
        """
        :return: the metadata keys that are common to all variables in the list
        """
        keys = list(self[0].metadata.keys())
        for var in self:
            keys = [key for key in var.metadata.keys() if key in keys]
        return keys

    def append(self, var: Variable) -> None:
        """
        Append var to the end of the list, unless its name is already used. In that case, var
        will replace the previous Variable instance with the same name.
        """
        if not isinstance(var, Variable):
            raise TypeError("VariableList items should be Variable instances")

        if var.name in self.names():
            self[self.names().index(var.name)] = var
        else:
            super().append(var)

    def update(self, other_var_list: list, add_variables: bool = True):
        """
        Uses variables in other_var_list to update the current VariableList instance.

        For each Variable instance in other_var_list:
            - if a Variable instance with same name exists, it is replaced by the one
              in other_var_list (special case: if one in other_var_list has an empty description,
              the original description is kept)
            - if not, Variable instance from other_var_list will be added only if
              add_variables==True

        :param other_var_list: source for new Variable data
        :param add_variables: if True, unknown variables are also added
        """

        for var in other_var_list:
            if add_variables or var.name in self.names():
                # To avoid to lose variables description when the variable list is updated with a
                # list without descriptions (issue # 319)
                if var.name in self.names() and self[var.name].description and not var.description:
                    var.description = self[var.name].description
                self.append(deepcopy(var))

    def to_ivc(self) -> om.IndepVarComp:
        """
        :return: an OpenMDAO IndepVarComp instance with all variables from current list
        """
        ivc = om.IndepVarComp()
        for variable in self:
            attributes = variable.metadata.copy()
            value = attributes.pop("val")
            # Some attributes are not compatible with add_output
            for attr in METADATA_TO_IGNORE:
                if attr in attributes:
                    del attributes[attr]
            ivc.add_output(variable.name, value, **attributes)

        return ivc

    def to_dataframe(self) -> pd.DataFrame:
        """
        Creates a DataFrame instance from a VariableList instance.

        Column names are "name" + the keys returned by :meth:`Variable.get_openmdao_keys`.
        Values in Series "value" are floats or lists (numpy arrays are converted).

        :return: a pandas DataFrame instance with all variables from current list
        """
        var_dict = {"name": []}
        var_dict.update({metadata_name: [] for metadata_name in self.metadata_keys()})

        for variable in self:
            value = self._as_list_or_float(variable.value)
            var_dict["name"].append(variable.name)
            for metadata_name in self.metadata_keys():
                if metadata_name == "val":
                    var_dict["val"].append(value)
                else:
                    # TODO: make this more generic
                    if metadata_name in ["val", "initial_value", "lower", "upper"]:
                        metadata = self._as_list_or_float(variable.metadata[metadata_name])
                    else:
                        metadata = variable.metadata[metadata_name]
                    var_dict[metadata_name].append(metadata)

        df = pd.DataFrame.from_dict(var_dict)

        return df

    @classmethod
    def from_dict(
        cls, var_dict: Union[Mapping[str, dict], Iterable[Tuple[str, dict]]]
    ) -> "VariableList":
        """
        Creates a VariableList instance from a dict-like object.

        :param var_dict:
        :return: a VariableList instance
        """
        variables = cls()

        for var_name, metadata in dict(var_dict).items():
            variables.append(Variable(var_name, **metadata))

        return variables

    @classmethod
    def from_ivc(cls, ivc: om.IndepVarComp) -> "VariableList":
        """
        Creates a VariableList instance from an OpenMDAO IndepVarComp instance

        :param ivc: an IndepVarComp instance
        :return: a VariableList instance
        """
        variables = cls()

        ivc = deepcopy(ivc)
        om.Problem(ivc).setup()  # Need setup to have get_io_metadata working

        for name, metadata in ivc.get_io_metadata(
            metadata_keys=["val", "units", "upper", "lower"]
        ).items():
            metadata = metadata.copy()
            value = metadata.pop("val")
            value = cls._as_list_or_float(value)
            metadata.update({"val": value})
            variables[name] = metadata

        return variables

    @classmethod
    def _as_list_or_float(cls, value):
        value = np.asarray(value)
        if np.size(value) == 1:
            value = value.item()
            try:
                value = float(value)
            except (TypeError, ValueError):
                pass
            return value

        return value.tolist()

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame) -> "VariableList":
        """
        Creates a VariableList instance from a pandas DataFrame instance.

        The DataFrame instance is expected to have column names "name" + some keys among the ones
        given by :meth:`Variable.get_openmdao_keys`.

        :param df: a DataFrame instance
        :return: a VariableList instance
        """
        column_names = [name for name in df.columns]

        def _get_variable(row):
            var_as_dict = {key: val for key, val in zip(column_names, row)}
            # TODO: make this more generic
            for key, val in var_as_dict.items():
                if key in ["val", "initial_value", "lower", "upper"]:
                    var_as_dict[key] = cls._as_list_or_float(val)
                else:
                    pass
            return Variable(**var_as_dict)

        return cls([_get_variable(row) for row in df[column_names].values])

    @classmethod
    def from_problem(
        cls,
        problem: om.Problem,
        use_initial_values: bool = False,
        get_promoted_names: bool = True,
        promoted_only: bool = True,
        io_status: str = "all",
    ) -> "VariableList":
        """
        Creates a VariableList instance containing inputs and outputs of an OpenMDAO Problem.

        The inputs (is_input=True) correspond to the variables of IndepVarComp
        components and all the unconnected variables.

        .. note::

            Variables from _auto_ivc are ignored.

        :param problem: OpenMDAO Problem instance to inspect
        :param use_initial_values: if True, or if problem has not been run, returned instance will
                                   contain values before computation
        :param get_promoted_names: if True, promoted names will be returned instead of absolute ones
                                   (if no promotion, absolute name will be returned)
        :param promoted_only: if True, only promoted variable names will be returned
        :param io_status: to choose with type of variable we return ("all", "inputs, "outputs")
        :return: VariableList instance
        """

        if not problem._metadata or problem._metadata["setup_status"] < _SetupStatus.POST_SETUP:
            with problem_without_mpi(problem) as problem_copy:
                problem_copy.setup()
                problem = problem_copy

        # Get inputs and outputs
        metadata_keys = (
            "val",
            "units",
            "shape",
            "size",
            "desc",
            "ref",
            "ref0",
            "lower",
            "upper",
            "tags",
        )
        inputs = problem.model.get_io_metadata("input", metadata_keys=metadata_keys)
        outputs = problem.model.get_io_metadata(
            "output", metadata_keys=metadata_keys, excludes="_auto_ivc.*"
        )
        indep_outputs = problem.model.get_io_metadata(
            "output", metadata_keys=metadata_keys, tags="indep_var"
        )
        # Move outputs from IndepVarComps into inputs
        for abs_name, metadata in indep_outputs.items():
            del outputs[abs_name]
            inputs[abs_name] = metadata

        # Remove non-promoted variables if needed
        if promoted_only:
            inputs = {
                name: metadata
                for name, metadata in inputs.items()
                if "." not in metadata["prom_name"]
            }
            outputs = {
                name: metadata
                for name, metadata in outputs.items()
                if "." not in metadata["prom_name"]
            }

            if get_promoted_names:
                # Check connections
                for name, metadata in inputs.copy().items():
                    source_name = problem.model.get_source(name)
                    if not (source_name.startswith("_auto_ivc.")) and source_name != name:
                        # This variable is connected to another variable of the problem: it is
                        # not an actual problem input. Let's move it to outputs.
                        del inputs[name]
                        outputs[name] = metadata

        # Add "is_input" field
        for metadata in inputs.values():
            metadata["is_input"] = True
        for metadata in outputs.values():
            metadata["is_input"] = False

        # Manage variable promotion
        if not get_promoted_names:
            final_inputs = inputs

            final_outputs = outputs
        else:
            final_inputs = {
                metadata["prom_name"]: dict(metadata, is_input=True) for metadata in inputs.values()
            }
            final_outputs = cls._get_promoted_outputs(outputs)

            # Remove possible duplicates due to Indeps
            for input_name in final_inputs:
                if input_name in final_outputs:
                    del final_outputs[input_name]

            # When variables are promoted, we may have retained a definition of the variable
            # that does not have any description, whereas a description is available in
            # another related definition (issue #319).
            # Therefore, we iterate again through original variable definitions to find
            # possible descriptions.
            for metadata in itertools.chain(inputs.values(), outputs.values()):
                prom_name = metadata["prom_name"]
                if not metadata["desc"]:
                    continue
                for final in final_inputs, final_outputs:
                    if prom_name in final and not final[prom_name]["desc"]:
                        final[prom_name]["desc"] = metadata["desc"]

        # Conversion to VariableList instances
        input_vars = cls.from_dict(final_inputs)
        output_vars = cls.from_dict(final_outputs)

        # Use computed value instead of initial ones, if asked for, and if problem has been run.
        # Note: using problem.get_val() if problem has not been run may lead to unexpected
        # behaviour when actually running the problem.
        if not use_initial_values and problem.model.iter_count > 0:
            for variable in input_vars + output_vars:
                try:
                    # Maybe useless, but we force units to ensure it is consistent
                    variable.value = problem.get_val(variable.name, units=variable.units)
                except RuntimeError:
                    # In case problem is incompletely set, problem.get_val() will fail.
                    # In such case, falling back to the method for initial values
                    # should be enough.
                    pass

        if io_status == "all":
            variables = input_vars + output_vars
        elif io_status == "inputs":
            variables = input_vars
        elif io_status == "outputs":
            variables = output_vars
        else:
            raise ValueError("Unknown value for io_status")

        return variables

    @classmethod
    def _get_promoted_outputs(cls, outputs: dict) -> dict:
        """

        :param outputs: dict (name, metadata) with non-promoted names as keys
        :return: dict (name, metadata) with promoted names as keys
        """
        promoted_outputs = {}
        for metadata in outputs.values():
            prom_name = metadata["prom_name"]
            # In case we get promoted names, several variables can match the same
            # promoted name, with possibly different declaration for default values.
            # We retain the first non-NaN value with defined units. If no units is
            # ever defined, the first non-NaN value is kept.
            # A non-NaN value with no units will be retained against a NaN value with
            # defined units.

            if prom_name in promoted_outputs:
                # prom_name has already been encountered.
                # Note: the succession of "if" is to help understanding, hopefully :)

                if not np.all(np.isnan(promoted_outputs[prom_name]["val"])):
                    if promoted_outputs[prom_name]["units"] is not None:
                        # We already have a non-NaN value with defined units for current
                        # promoted name. No need for using the current variable.
                        continue
                    if np.all(np.isnan(metadata["val"])):
                        # We already have a non-NaN value and current variable has a NaN value,
                        # so it can only add information about units. We keep the non-NaN value
                        continue

                if (
                    np.all(np.isnan(promoted_outputs[prom_name]["val"]))
                    and metadata["units"] is None
                ):
                    # We already have a non-NaN value and current variable provides no unit.
                    # No need for using the current variable.
                    continue
            promoted_outputs[prom_name] = metadata

        return promoted_outputs

    @classmethod
    @deprecated(
        version="1.3.0",
        reason="Will be removed in version 2.0. Please use VariableList.from_problem() instead",
    )
    def from_unconnected_inputs(
        cls, problem: om.Problem, with_optional_inputs: bool = False
    ) -> "VariableList":
        """
        Creates a VariableList instance containing unconnected inputs of an OpenMDAO Problem.

        .. warning::

            problem.setup() must have been run.

        If *optional_inputs* is False, only inputs that have numpy.nan as default value (hence
        considered as mandatory) will be in returned instance. Otherwise, all unconnected inputs
        will be in returned instance.

        :param problem: OpenMDAO Problem instance to inspect
        :param with_optional_inputs: If True, returned instance will contain all unconnected inputs.
                                Otherwise, it will contain only mandatory ones.
        :return: VariableList instance
        """
        variables = VariableList()

        mandatory_unconnected, optional_unconnected = get_unconnected_input_names(problem)
        model = problem.model

        # processed_prom_names will store promoted names that have been already processed, so that
        # it won't be stored twice.
        # By processing mandatory variable first, a promoted variable that would be mandatory
        # somewhere and optional elsewhere will be retained as mandatory (and associated value
        # will be NaN), which is fine.
        # For promoted names that link to several optional variables and no mandatory ones,
        # associated value will be the first encountered one, and this is as good a choice as any
        # other.
        processed_prom_names = []

        io_metadata = model.get_io_metadata(
            metadata_keys=["val", "units", "desc"], return_rel_names=False
        )

        def _add_outputs(unconnected_names):
            """Fills ivc with data associated to each provided var"""
            for abs_name in unconnected_names:
                prom_name = io_metadata[abs_name]["prom_name"]
                if prom_name not in processed_prom_names:
                    processed_prom_names.append(prom_name)
                    metadata = deepcopy(io_metadata[abs_name])
                    metadata.update({"is_input": True})
                    variables[prom_name] = metadata
                elif not variables[prom_name].description and io_metadata[abs_name]["desc"]:
                    variables[prom_name].description = io_metadata[abs_name]["desc"]

        _add_outputs(mandatory_unconnected)
        if with_optional_inputs:
            _add_outputs(optional_unconnected)

        return variables

    def __getitem__(self, key) -> Variable:
        if isinstance(key, str):
            return self[self.names().index(key)]
        else:
            return super().__getitem__(key)

    def __setitem__(self, key, value):
        if isinstance(key, str):
            if isinstance(value, dict):
                variable = Variable(key, **value)
                if key in self.names():
                    self[key].metadata = variable.metadata
                else:
                    self.append(variable)
            else:
                raise TypeError(
                    'VariableList can be set with "vars[key] = value" only if value is a '
                    "dict of metadata"
                )
        elif not isinstance(value, Variable):
            raise TypeError("VariableList items should be Variable instances")
        else:
            super().__setitem__(key, value)

    def __delitem__(self, key):
        if isinstance(key, str):
            del self[self.names().index(key)]
        else:
            super().__delitem__(key)

    def __add__(self, other) -> Union[List, "VariableList"]:
        if isinstance(other, VariableList):
            return VariableList(super().__add__(other))
        else:
            return super().__add__(other)



"""
Classes for managing internal structures of missions.

The mission file provides a "human" definition of the mission.
Structures are the translation of this human definition, that is ready to
be transformed in a Python implementation.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from collections import OrderedDict
from copy import deepcopy
from dataclasses import InitVar, dataclass, field, fields
from itertools import chain
from typing import List, Tuple

import numpy as np

from .constants import NAME_TAG, SEGMENT_TYPE_TAG, TYPE_TAG
from .input_definition import InputDefinition
from ..schema import (
    CLIMB_PARTS_TAG,
    CRUISE_PART_TAG,
    DESCENT_PARTS_TAG,
    MISSION_DEFINITION_TAG,
    PARTS_TAG,
    PHASE_DEFINITIONS_TAG,
    PHASE_TAG,
    ROUTE_DEFINITIONS_TAG,
    ROUTE_TAG,
    SEGMENT_TAG,
)
from ...segments.base import SegmentDefinitions


@dataclass
class AbstractStructureBuilder(ABC):
    """
    Base class for building structures of mission parts.

    "Structures" are dicts that are derived from the mission definition file so that they can be
    readily translated into the matching implementation.

    Usage:

    Subclasses must implement the build method that will create the specific part of the structure
    dict (name and type fields are automated).

    If the structure has to contain the result of another result, :meth:`insert_builder` should
    be used to ensure a correct processing of the global structure, especially to get a correct
    resolution of :attr:`input_definitions`.
    """

    definition: InitVar[dict]
    name: str
    parent_name: str = None

    _structure: OrderedDict = field(default=None, init=False)

    _input_definitions: List[InputDefinition] = field(default_factory=list, init=False)
    _builders: List[Tuple["AbstractStructureBuilder", dict]] = field(
        default_factory=list, init=False
    )

    #: Defined by subclass
    type = None

    def __init_subclass__(cls, *, structure_type=None):
        cls.type = structure_type

    def __post_init__(self, definition):
        self._structure = self._build(definition)
        self._structure[NAME_TAG] = self.qualified_name
        if self.__class__.type:
            self._structure[TYPE_TAG] = self.__class__.type
        self._parse_inputs(self._structure, self._input_definitions)

    @property
    def structure(self) -> OrderedDict:
        """A dictionary that is ready to be translated to the matching implementation."""
        for builder, placeholder in self._builders:
            placeholder.update(builder.structure)
            self._input_definitions += builder.get_input_definitions()
        self._builders = []  # Builders have been used and can be forgotten.
        return self._structure

    def get_input_definitions(self) -> List[InputDefinition]:
        """List of InputDefinition instances in the structure."""
        return self._input_definitions + list(
            chain(*[builder.get_input_definitions() for builder, _ in self._builders])
        )

    def process_builder(self, builder: "AbstractStructureBuilder") -> dict:
        """
        Method to be used when another StructureBuilder object should be inserted in
        :attr:`structure`.

        Not using this method will prevent a correct processing of :attr:`input_definitions`.

        .. note::

            The returned object is always an empty dict. It is actually a memory reference that
            will allow to fill this "placeholder" later with the final result of the builder, that
            cannot be completely known when builder is created from read definition.

        :param builder: the builder object
        :return: the object that has to be put at location where the builder result should be used
        """

        placeholder = {}
        self._builders.append((builder, placeholder))
        return placeholder

    @abstractmethod
    def _build(self, definition: dict) -> OrderedDict:
        """
        This method creates the needed structure dict.

        Keys "name" and "type" are not needed, as they will be written later on in the process.

        .. Important::

            Please use :meth:`_insert_builder` when another StructureBuilder object is needed in
            the currently built structure.

        :param definition: the dict that will be converted.
        :return: the structure dict
        """

    @property
    def qualified_name(self):
        """Name of the current structure, preceded by the parent names, separated by colons (:)."""
        if not self.parent_name:
            return self.name

        name = self.parent_name
        if self.name:
            name += f":{self.name}"
        return name

    def _parse_inputs(
        self,
        structure,
        input_definitions: List[InputDefinition],
        parent=None,
        part_identifier="",
        segment_class=None,
    ):
        """
        Returns the `definition` structure where all inputs (string/numeric values, numeric lists,
        dicts with a "value key"), have been converted to an InputDefinition instance.

        Created InputDefinition instances are stored in provided input_definitions.

        :param structure:
        :param parent:
        :param part_identifier:
        :return: amended definition
        """

        if isinstance(structure, dict):
            if "polar" in structure:
                structure["polar"] = self._process_polar(structure["polar"])

            if "value" in structure:
                input_definition = InputDefinition.from_dict(
                    parent, structure, part_identifier=part_identifier
                )
                input_definitions.append(input_definition)
                return input_definition

            part_identifier = structure.get(NAME_TAG, part_identifier)
            segment_class = SegmentDefinitions.get_segment_class(structure.get(SEGMENT_TYPE_TAG))
            for key, value in structure.items():
                if key in [
                    NAME_TAG,
                    TYPE_TAG,
                    SEGMENT_TYPE_TAG,
                    PARTS_TAG,
                    CLIMB_PARTS_TAG,
                    DESCENT_PARTS_TAG,
                    CRUISE_PART_TAG,
                ]:
                    continue
                if (
                    segment_class
                    and isinstance(value, dict)
                    and isinstance(value.get("value"), str)
                    and self._is_shape_by_conn(key, segment_class)
                ):
                    value["shape_by_conn"] = True

                structure[key] = self._parse_inputs(
                    value,
                    input_definitions,
                    parent=key,
                    part_identifier=part_identifier,
                    segment_class=segment_class,
                )
            return structure

        # Here structure is not a dict, hence a directly given value.
        key, value = parent, structure
        input_definition = InputDefinition(key, value, part_identifier=part_identifier)
        if segment_class and isinstance(value, str):
            input_definition.shape_by_conn = self._is_shape_by_conn(key, segment_class)
        input_definitions.append(input_definition)
        return input_definition

    @staticmethod
    def _is_shape_by_conn(key, segment_class) -> bool:
        """
        Here variables that are expected to be arrays or lists in the provided segment class are
        attributed the "shape_by_conn=True" property.
        """
        segment_fields = [fld for fld in fields(segment_class) if fld.name == key]
        return len(segment_fields) == 1 and issubclass(segment_fields[0].type, (list, np.ndarray))

    @staticmethod
    def _process_polar(polar_structure):
        """
        Polar data are handled specifically, as it a particular parameter of segments (
        a Polar class).

        If "foo:bar:baz" is provided as `polar_structure`, it is replaced by the dict
        { "CL": {value:"foo:bar:baz:CL", "shape_by_conn": True},
          "CD": {value:"foo:bar:baz:CD", "shape_by_conn": True}}

        Also, even if `polar_structure` is a dict, it is ensured that it has the structure above.
        """

        if isinstance(polar_structure, str):
            polar_structure = OrderedDict(
                {
                    "CL": {"value": polar_structure + ":CL", "shape_by_conn": True},
                    "CD": {"value": polar_structure + ":CD", "shape_by_conn": True},
                }
            )
        elif isinstance(polar_structure, dict):
            for key in ["CL", "CD"]:
                if isinstance(polar_structure[key], str):
                    polar_structure[key] = {"value": polar_structure[key], "shape_by_conn": True}
                elif isinstance(polar_structure[key], dict):
                    polar_structure[key]["shape_by_conn"] = True

        return polar_structure


class DefaultStructureBuilder(AbstractStructureBuilder):
    """
    Builder for structures that do not need to process the given definition.

    :param definition: the definition for the part only
    """

    def _build(self, definition: dict) -> OrderedDict:
        return OrderedDict(deepcopy(definition))


class SegmentStructureBuilder(AbstractStructureBuilder, structure_type=SEGMENT_TAG):
    """
    Structure builder for segment definition.

    :param definition: the definition for the segment only
    """

    def _build(self, definition: dict) -> OrderedDict:
        segment_structure = OrderedDict(deepcopy(definition))
        del segment_structure[SEGMENT_TAG]
        segment_structure[SEGMENT_TYPE_TAG] = definition[SEGMENT_TAG]

        return segment_structure


class PhaseStructureBuilder(AbstractStructureBuilder, structure_type=PHASE_TAG):
    """
    Structure builder for phase definition.

    :param definition: the whole content of definition file
    """

    def _build(self, definition: dict) -> OrderedDict:
        phase_definition = definition[PHASE_DEFINITIONS_TAG][self.name]
        phase_structure = OrderedDict(deepcopy(phase_definition))

        for i, part in enumerate(phase_definition[PARTS_TAG]):
            if PHASE_TAG in part:
                builder = PhaseStructureBuilder(definition, part[PHASE_TAG], self.qualified_name)
            elif SEGMENT_TAG in part:
                builder = SegmentStructureBuilder(part, "", self.qualified_name)
            else:
                raise RuntimeError(f"Unexpected structure in definition of phase {self.name}")

            phase_structure[PARTS_TAG][i] = {}
            phase_structure[PARTS_TAG][i] = self.process_builder(builder)

        return phase_structure


class RouteStructureBuilder(AbstractStructureBuilder, structure_type=ROUTE_TAG):
    """
    Structure builder for route definition.

    :param definition: the whole content of definition file
    """

    def _build(self, definition: dict) -> OrderedDict:
        route_definition = definition[ROUTE_DEFINITIONS_TAG][self.name]
        route_structure = OrderedDict(deepcopy(route_definition))

        route_structure[CLIMB_PARTS_TAG] = self._get_route_climb_or_descent_structure(
            definition, route_definition[CLIMB_PARTS_TAG]
        )

        builder = SegmentStructureBuilder(
            route_definition[CRUISE_PART_TAG], "cruise", self.qualified_name
        )
        route_structure[CRUISE_PART_TAG] = self.process_builder(builder)

        route_structure[DESCENT_PARTS_TAG] = self._get_route_climb_or_descent_structure(
            definition, route_definition[DESCENT_PARTS_TAG]
        )

        return route_structure

    def _get_route_climb_or_descent_structure(self, global_definition, parts_definition):
        parts = []
        for part_definition in parts_definition:
            phase_name = part_definition["phase"]
            builder = PhaseStructureBuilder(global_definition, phase_name, self.qualified_name)
            phase_structure = self.process_builder(builder)
            parts.append(phase_structure)
        return parts


class MissionStructureBuilder(AbstractStructureBuilder, structure_type="mission"):
    """
    Structure builder for mission definition.

    :param definition: the whole content of definition file
    """

    def _build(self, definition: dict) -> OrderedDict:
        mission_definition = definition[MISSION_DEFINITION_TAG][self.name]
        mission_structure = OrderedDict(deepcopy(mission_definition))

        mission_parts = []
        for part_definition in mission_definition[PARTS_TAG]:
            if ROUTE_TAG in part_definition:
                route_name = part_definition[ROUTE_TAG]
                builder = RouteStructureBuilder(definition, route_name, self.qualified_name)
            elif PHASE_TAG in part_definition:
                phase_name = part_definition[PHASE_TAG]
                builder = PhaseStructureBuilder(definition, phase_name, self.qualified_name)
            elif SEGMENT_TAG in part_definition:
                builder = SegmentStructureBuilder(part_definition, "", self.qualified_name)
            else:
                builder = DefaultStructureBuilder(part_definition, "", self.qualified_name)

            part_structure = self.process_builder(builder)
            mission_parts.append(part_structure)

        mission_structure[PARTS_TAG] = mission_parts

        return mission_structure

    @property
    def qualified_name(self):
        return self.name



"""Class for specifying input mass at "any" point in the mission."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass

import pandas as pd

from fastoad.model_base import FlightPoint
from fastoad.models.performances.mission.segments.base import AbstractFlightSegment


@dataclass
class MassTargetSegment(AbstractFlightSegment, mission_file_keyword="mass_input"):
    """
    Class that simply sets a target mass.

    :meth:`compute_from` returns a 1-row dataframe that is the start point with mass
    set to provided target mass.

    class:`~fastoad.models.performances.mission.base.FlightSequence` ensures that
    mass is consistent for segments prior to this one.
    """

    def compute_from_start_to_target(self, start: FlightPoint, target: FlightPoint) -> pd.DataFrame:
        start.mass = target.mass
        self.complete_flight_point(start)
        return pd.DataFrame([start])



"""
Mission wrapper.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Tuple

import numpy as np
import openmdao.api as om
import pandas as pd
from openmdao.vectors.vector import Vector

from fastoad.model_base import FlightPoint
from ..mission_definition.mission_builder import MissionBuilder
from ..mission_definition.schema import (
    CLIMB_PARTS_TAG,
    DESCENT_PARTS_TAG,
    MISSION_DEFINITION_TAG,
    PARTS_TAG,
    PHASE_TAG,
    RESERVE_TAG,
    ROUTE_DEFINITIONS_TAG,
    ROUTE_TAG,
)


class MissionWrapper(MissionBuilder):
    """
    Wrapper around
    :class:`~fastoad.models.performances.mission.mission_definition.mission_builder.MissionBuilder`
    for using with OpenMDAO.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mission_name = None
        self.consumed_fuel_before_input_weight = 0.0

    def setup(self, component: om.ExplicitComponent, mission_name: str = None):
        """
        To be used during setup() of provided OpenMDAO component.

        It adds input and output variables deduced from mission definition file.

        :param component: the OpenMDAO component where the setup is done.
        :param mission_name: mission name (can be omitted if only one mission is defined)
        """

        if mission_name is None:
            mission_name = self.get_unique_mission_name()
        self.mission_name = mission_name
        input_definition = self.get_input_variables(mission_name)
        output_definition = self._identify_outputs()
        output_definition = {
            name: value
            for name, value in output_definition.items()
            if name not in input_definition.names()
        }
        for variable in input_definition:
            component.add_input(**variable.get_openmdao_kwargs())

        for name, (units, desc) in output_definition.items():
            component.add_output(name, 0.0, units=units, desc=desc)

    def compute(
        self, start_flight_point: FlightPoint, inputs: Vector, outputs: Vector
    ) -> pd.DataFrame:
        """
        To be used during compute() of an OpenMDAO component.

        Builds the mission from input file, and computes it. `outputs` vector is
        filled with duration, burned fuel and covered ground distance for each
        part of the flight.

        :param start_flight_point: starting point of mission
        :param inputs: the input vector of the OpenMDAO component
        :param outputs: the output vector of the OpenMDAO component
        :return: a pandas DataFrame where column names match fields of
                 :class:`~fastoad.model_base.flight_point.FlightPoint`
        """
        mission = self.build(inputs, self.mission_name)

        def _compute_vars(name_root, start: FlightPoint, end: FlightPoint):
            """Computes duration, burned fuel and covered distance."""
            if name_root + ":duration" in outputs:
                outputs[name_root + ":duration"] = end.time - start.time
            if name_root + ":fuel" in outputs:
                outputs[name_root + ":fuel"] = start.mass - end.mass
            if name_root + ":distance" in outputs:
                outputs[name_root + ":distance"] = end.ground_distance - start.ground_distance

        flight_points = mission.compute_from(start_flight_point)
        flight_points.loc[0, "name"] = flight_points.loc[1, "name"]

        nb_levels = np.max([len(n.split(":")) for n in flight_points["name"]])
        for i in range(nb_levels):
            flight_points["name2"] = [":".join(n.split(":")[: i + 1]) for n in flight_points.name]
            grouped_points = flight_points.groupby("name2")

            part_names = pd.unique(flight_points.name2)
            for part_name1, part_name2 in zip(part_names[:-1], part_names[1:]):
                part1 = grouped_points.get_group(part_name1)
                part2 = grouped_points.get_group(part_name2)
                _compute_vars(f"data:mission:{part_name2}", part1.iloc[-1], part2.iloc[-1])

            start_part_name = part_names[0]
            start_part = grouped_points.get_group(start_part_name)
            _compute_vars(
                f"data:mission:{start_part_name}", start_part.iloc[0], start_part.iloc[-1]
            )

        del flight_points["name2"]
        self.consumed_fuel_before_input_weight = mission.consumed_mass_before_input_weight
        return flight_points

    def get_reserve_variable_name(self) -> str:
        """
        :return: the name of OpenMDAO variable for fuel reserve. This name is among the declared
                 outputs in :meth:`setup`.
        """
        return f"{self._variable_prefix}:reserve:fuel"

    def _identify_outputs(self) -> Dict[str, Tuple[str, str]]:
        """
        Builds names of OpenMDAO outputs from names of mission, route and phases.

        :return: dictionary with variable name as key and unit, description as value
        """
        output_definition = {}

        output_definition.update(self._add_vars(self.mission_name))

        for part in self.definition[MISSION_DEFINITION_TAG][self.mission_name][PARTS_TAG]:
            if PHASE_TAG in part:
                phase_name = part[PHASE_TAG]
                output_definition.update(self._add_vars(self.mission_name, phase_name=phase_name))
            elif ROUTE_TAG in part:
                route_name = part[ROUTE_TAG]
                output_definition.update(self._add_vars(self.mission_name, route_name))
                route_definition = self.definition[ROUTE_DEFINITIONS_TAG][route_name]
                for part_definition in list(
                    route_definition[CLIMB_PARTS_TAG] + route_definition[DESCENT_PARTS_TAG]
                ):
                    phase_name = part_definition[PHASE_TAG]
                    output_definition.update(
                        self._add_vars(self.mission_name, route_name, phase_name)
                    )
                output_definition.update(self._add_vars(self.mission_name, route_name, "cruise"))
            elif RESERVE_TAG in part:
                output_definition[self.get_reserve_variable_name()] = (
                    "kg",
                    f'reserve fuel for mission "{self.mission_name}"',
                )

        return output_definition

    @staticmethod
    def _add_vars(mission_name, route_name=None, phase_name=None) -> dict:
        """
        Builds names of OpenMDAO outputs for provided mission, route and phase names.

        :param mission_name:
        :param route_name:
        :param phase_name:
        :return: dictionary with variable name as key and unit, description as value
        """
        output_definition = {}

        name_root = ":".join(
            name for name in ["data:mission", mission_name, route_name, phase_name] if name
        )
        if route_name and phase_name:
            flight_part_desc = (
                f'phase "{phase_name}" of route "{route_name}" in mission "{mission_name}"'
            )

        elif route_name:
            flight_part_desc = f'route "{route_name}" in mission "{mission_name}"'
        elif phase_name:
            flight_part_desc = f'phase "{phase_name}" in mission "{mission_name}"'
        else:
            flight_part_desc = f'mission "{mission_name}"'

        output_definition[name_root + ":duration"] = ("s", f"duration of {flight_part_desc}")
        output_definition[name_root + ":fuel"] = ("kg", f"burned fuel during {flight_part_desc}")
        output_definition[name_root + ":distance"] = (
            "m",
            f"covered ground distance during {flight_part_desc}",
        )

        return output_definition

    @property
    def _variable_prefix(self):
        return f"data:mission:{self.mission_name}"



"""
Basis for registering and retrieving services
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import gc
import logging
import re
from typing import Any, Dict, List, Optional, Set, Tuple, Type, TypeVar, Union

import pelix
from pelix.constants import BundleException
from pelix.framework import Bundle, BundleContext, Framework, FrameworkFactory
from pelix.internals.registry import ServiceReference
from pelix.ipopo.constants import SERVICE_IPOPO, use_ipopo
from pelix.ipopo.decorators import ComponentFactory, Property, Provides

from .exceptions import (
    FastBundleLoaderDuplicateFactoryError,
    FastBundleLoaderUnknownFactoryNameError,
)
from .._utils.resource_management.contents import PackageReader

_LOGGER = logging.getLogger(__name__)
"""Logger for this module"""

T = TypeVar("T")


class BundleLoader:
    """
    Helper class for loading Pelix bundles.
    """

    def __init__(self):
        """
        Constructor
        """
        self._framework = None

    @property
    def framework(self) -> Framework:
        """
        The currently running Pelix framework instance, or a new one if none is running (anyway,
        the framework instance will continue after deletion of this BundleLoader instance)
        """
        if FrameworkFactory.is_framework_running():
            self._framework = FrameworkFactory.get_framework()
            # Do not use self.context, because it will call back this.
            context = self._framework.get_bundle_context()
            if not context.get_service_reference(SERVICE_IPOPO):
                self._framework.install_bundle(SERVICE_IPOPO)
        else:
            self._framework = pelix.framework.create_framework(())
            self._framework.install_bundle(SERVICE_IPOPO)
            self._framework.start()
        return self._framework

    @property
    def context(self) -> BundleContext:
        """
        BundleContext instance of running Pelix framework
        """
        return self.framework.get_bundle_context()

    def explore_folder(
        self, folder_path: str, is_package: bool = False
    ) -> Tuple[Set[Bundle], Set[str]]:
        """
        Installs bundles found in *folder_path*.

        :param folder_path: The path of folder to scan
        :param is_package: if True, folder_path should be a Python package name that will be
                           scanned using importlib.resources
        :return: A 2-tuple, with the list of installed bundles
                 (:class:`~pelix.framework.Bundle`) and the list of the names
                 of the modules which import failed.
        """

        _LOGGER.info("Loading bundles from %s", folder_path)

        if is_package:
            bundles, failed = self._install_python_package(folder_path)
        else:
            bundles, failed = self.framework.install_package(folder_path, True)

        for bundle in bundles:
            _LOGGER.info(
                "Installed bundle %s (ID %s )", bundle.get_symbolic_name(), bundle.get_bundle_id()
            )
            bundle.start()
        for _f in failed:
            _LOGGER.warning("Failed to import module %s", _f)

        return bundles, failed

    def get_services(
        self, service_name: str, properties: dict = None, case_sensitive: bool = False
    ) -> Optional[list]:
        """
        Returns the services that match *service_name* and provided
        *properties* (if provided).

        :param service_name:
        :param properties:
        :param case_sensitive: if False, case of property values will be
                               ignored
        :return: the list of service instances
        """
        references = self._get_service_references(service_name, properties, case_sensitive)
        services = None
        if references is not None:
            services = [self.context.get_service(ref) for ref in references]

        return services

    def register_factory(
        self,
        component_class: Type[T],
        factory_name: str,
        service_names: Union[List[str], str],
        properties: dict = None,
    ) -> Type[T]:
        """
        Registers provided class as iPOPO component factory.

        :param component_class: the class of the components that will be provided by the factory
        :param factory_name: the name of the factory
        :param service_names: the service(s) that will be provided by the components
        :param properties: the properties associated to the factory
        :return: the input class, amended by iPOPO
        :raise FastDuplicateFactoryError:
        """

        obj = Provides(service_names)(component_class)
        with use_ipopo(self.context) as ipopo:
            if ipopo.is_registered_factory(factory_name):
                raise FastBundleLoaderDuplicateFactoryError(factory_name)

        if properties:
            for key, value in properties.items():
                obj = Property(field="_" + self._fieldify(key), name=key, value=value)(obj)

        factory = ComponentFactory(factory_name)(obj)

        return factory

    def get_factory_names(
        self, service_name: str = None, properties: dict = None, case_sensitive: bool = False
    ) -> List[str]:
        """
        Browses the available factory names to find what factories provide `service_name` (if
        provided) and match provided `properties` (if provided).

        if neither service_name nor properties are provided, all registered factory
        names are returned.

        :param service_name:
        :param properties:
        :param case_sensitive: if False, case of property values will be ignored
        :return: the list of factory names
        """
        with use_ipopo(self.context) as ipopo:
            all_names = ipopo.get_factories()
            if not service_name and not properties:
                return all_names

            names = []
            for name in all_names:
                details = ipopo.get_factory_details(name)
                to_be_kept = True
                if service_name and service_name not in details["services"][0]:
                    to_be_kept = False
                elif properties:
                    factory_properties: dict = details["properties"]
                    if case_sensitive:
                        to_be_kept = all(
                            item in factory_properties.items() for item in properties.items()
                        )
                    else:
                        for prop_name, prop_value in properties.items():
                            if prop_name not in factory_properties.keys():
                                to_be_kept = False
                                break
                            factory_prop_value = factory_properties[prop_name]
                            if isinstance(prop_value, str):
                                prop_value = prop_value.lower()
                                factory_prop_value = factory_prop_value.lower()
                            if prop_value != factory_prop_value:
                                to_be_kept = False
                                break
                if to_be_kept:
                    names.append(name)
        return names

    def get_factory_path(self, factory_name: str) -> str:
        """

        :param factory_name:
        :return: path of the file where the factory is defined
        """

        details = self.get_factory_details(factory_name)
        bundle: Bundle = details["bundle"]
        return bundle.get_location()

    def get_factory_properties(self, factory_name: str) -> dict:
        """
        :param factory_name:
        :return: properties of the factory
        """

        details = self.get_factory_details(factory_name)
        properties = details["properties"]
        return properties

    def get_factory_property(self, factory_name: str, property_name: str) -> Any:
        """
        :param factory_name:
        :param property_name:
        :return: property value, or None if property is not found
        """

        properties = self.get_factory_properties(factory_name)
        return properties.get(property_name)

    def get_factory_details(self, factory_name: str) -> Dict[str, Any]:
        """
        :param factory_name: name of the factory
        :return: factory details as in iPOPO
        :raise FastBundleLoaderUnknownFactoryNameError: unknown factory name
        """
        with use_ipopo(self.context) as ipopo:
            try:
                return ipopo.get_factory_details(factory_name)
            except ValueError as exc:
                raise FastBundleLoaderUnknownFactoryNameError(factory_name) from exc

    def get_instance_property(self, instance: Any, property_name: str) -> Any:
        """
        :param instance: any instance from :meth:~BundleLoader.instantiate_component
        :param property_name:
        :return: property value, or None if property is not found
        """

        try:
            return getattr(instance, "_" + self._fieldify(property_name))
        except AttributeError:
            return None

    def instantiate_component(self, factory_name: str, properties: dict = None) -> Any:
        """
        Instantiates a component from given factory

        :param factory_name: name of the factory
        :param properties: Initial properties of the component instance
        :return: the component instance
        :raise FastBundleLoaderUnknownFactoryNameError: unknown factory name
        """
        with use_ipopo(self.context) as ipopo:
            try:
                return ipopo.instantiate(
                    factory_name, self._get_instance_name(factory_name), properties
                )
            except TypeError as exc:
                raise FastBundleLoaderUnknownFactoryNameError(factory_name) from exc

    def clean_memory(self):
        """
        Removes all service objects from memory and runs garbage collector.
        """
        for bundle in self.context.get_bundles():
            # FIXME: it would be better to target bundles that provide FAST-OAD registration
            #        instead of just avoiding iPOPO bundles.
            if bundle.get_state() > Bundle.INSTALLED and "ipopo" not in bundle.get_symbolic_name():
                bundle.stop()
                bundle.start()
        gc.collect()

    def _get_instance_name(self, base_name: str):
        """
        Creates an instance name that is not currently used by iPOPO

        :param base_name: the beginning of the instance name
        :return: the created instance name
        """
        with use_ipopo(self.context) as ipopo:
            instances = ipopo.get_instances()
            instance_names = [i[0] for i in instances]
            i = 0
            name = "%s_%i" % (base_name, i)
            while name in instance_names:
                i = i + 1
                name = "%s_%i" % (base_name, i)

            return name

    def _get_service_references(
        self, service_name: str, properties: dict = None, case_sensitive: bool = False
    ) -> Optional[List[ServiceReference]]:
        """
        Returns the service references that match *service_name* and
        provided *properties* (if provided)

        :param service_name:
        :param properties:
        :param case_sensitive: if False, case of property values will be
                               ignored
        :return: a list of ServiceReference instances
        """

        # Dev Note: simple wrapper for BundleContext.get_all_service_references()

        if case_sensitive:
            operator = "="
        else:
            operator = "~="

        if not properties:
            ldap_filter = None
        else:
            ldap_filter = (
                "(&"
                + "".join(
                    [
                        "({0}{1}{2})".format(key, operator, value)
                        for key, value in properties.items()
                    ]
                )
                + ")"
            )
            _LOGGER.debug(ldap_filter)

        references = self.framework.find_service_references(service_name, ldap_filter)
        return references

    def _install_python_package(self, package_name: str) -> Tuple[Set[Bundle], Set[str]]:
        """
        Recursively loads indicated package.

        :param package_name:
        :return: A 2-tuple, with the list of installed bundles and the list
                 of failed modules names
        """
        bundles = set()
        failed = set()

        package = PackageReader(package_name)
        if package.has_error or not package.exists:
            failed.add(package_name)
        elif package.is_package:
            # It is a package, let's explore it.
            for item in package.contents:
                item_package = ".".join([package_name, item])
                if "." in item:
                    # A file. Considered only if it is a Python file. Ignored otherwise.
                    if item.endswith(".py"):
                        try:
                            bundle = self.context.install_bundle(item_package[:-3])
                            bundles.add(bundle)
                        except BundleException:
                            failed.add(package_name)
                else:
                    sub_bundles, sub_failed = self._install_python_package(item_package)
                    bundles.update(sub_bundles)
                    failed.update(sub_failed)

        return bundles, failed

    @staticmethod
    def _fieldify(name: str) -> str:
        """
        Converts a string into a valid field name, i.e. replaces all spaces and
        non-word characters with an underscore.

        :param name: the string to fieldify
        :return: the field version of `name`
        """
        return re.compile(r"[\W_]+").sub("_", name).strip("_")



def compute_cd0(wing_area):
    """
    Computes the profile drag of the aircraft based on the wing area

    :param wing_area: Wing area, in m2

    :return cd0: profile drag, no unit
    """

    # Wet area of the aircraft without the wings
    wing_area_ref = 13.50

    # Profile drag coefficient of the aircraft without the wings
    cd0_other = 0.022

    # Constant linking the wing profile drag to its wet area, and by extension, its reference area
    c = 0.0004

    # Computation of the profile drag
    cd0 = cd0_other * wing_area_ref / wing_area + c

    return cd0



"""
Convenience functions for helping tests
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import numpy as np
import openmdao.api as om
from openmdao.core.system import System

from fastoad.openmdao.variables import VariableList

_LOGGER = logging.getLogger(__name__)  # Logger for this module


def run_system(
    component: System, input_vars: om.IndepVarComp, setup_mode="auto", add_solvers=False
):
    """Runs and returns an OpenMDAO problem with provided component and data"""
    problem = om.Problem()
    model = problem.model
    model.add_subsystem("inputs", input_vars, promotes=["*"])
    model.add_subsystem("component", component, promotes=["*"])
    if add_solvers:
        model.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        model.linear_solver = om.DirectSolver()

    problem.setup(mode=setup_mode)
    variable_names = [
        var.name
        for var in VariableList.from_problem(problem, io_status="inputs")
        if np.any(np.isnan(var.val))
    ]

    assert not variable_names, "These inputs are not provided: %s" % variable_names

    problem.run_model()

    return problem



import scipy.constants as sc


def compute_wing_mass(mtow, aspect_ratio, wing_area):
    """
    Computes the wing mass based on the MTOW, its area and aspect ratio

    :param wing_area: Wing area, in m2
    :param aspect_ratio: Wing aspect ratio, no unit
    :param mtow: Max Take-Off Weight, in kg

    :return wing mass: the wing_mass, in kg
    """

    # Let's start by converting the quantities in imperial units
    mtow_lbm = mtow / sc.lb
    wing_area_ft2 = wing_area / sc.foot ** 2.0

    # Let's now apply the formula
    wing_mass_lbm = (
        96.948
        * (
            (5.7 * mtow_lbm / 1.0e5) ** 0.65
            * aspect_ratio ** 0.57
            * (wing_area_ft2 / 100.0) ** 0.61
            * 2.5
        )
        ** 0.993
    )

    # Converting wing mass in kg
    wing_mass = wing_mass_lbm * sc.lb

    return wing_mass



from .sub_components.compute_wing_mass import compute_wing_mass
from .sub_components.compute_owe import compute_owe


def compute_mass(mtow, wing_area, aspect_ratio):
    """
    Gather all the mass sub-functions in the main function

    :param mtow: Max Take-Off Weight, in kg
    :param wing_area: Wing area, in m2
    :param aspect_ratio: Wing aspect ratio, no unit

    :return owe: the structural mass, in kg
    """

    # Let's start by computing the wing mass
    wing_mass = compute_wing_mass(mtow=mtow, aspect_ratio=aspect_ratio, wing_area=wing_area)

    # Let's now compute the owe
    owe = compute_owe(
        wing_mass=wing_mass,
        mtow=mtow,
    )

    return owe



"""Classes for simulating cruise segments."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from copy import deepcopy
from dataclasses import dataclass
from typing import List

import numpy as np
import pandas as pd
from scipy.constants import foot, g

from fastoad.model_base import FlightPoint
from .altitude_change import AltitudeChangeSegment
from .base import AbstractRegulatedThrustSegment, AbstractTimeStepFlightSegment
from ..util import get_closest_flight_level


@dataclass
class CruiseSegment(AbstractRegulatedThrustSegment):
    """
    Class for computing cruise flight segment at constant altitude and speed.

    Mach is considered constant, equal to Mach at starting point.
    Altitude is constant.
    Target is a specified ground_distance. The target definition indicates
    the ground_distance to be covered during the segment, independently of
    the initial value.
    """

    def __post_init__(self):
        super().__post_init__()
        # Constant speed at constant altitude is necessarily constant Mach, but
        # subclasses can be at variable altitude, so Mach is considered constant
        # if no other constant speed parameter is set to "constant".
        if AbstractTimeStepFlightSegment.CONSTANT_VALUE not in [
            self.target.true_airspeed,
            self.target.equivalent_airspeed,
        ]:
            self.target.mach = AbstractTimeStepFlightSegment.CONSTANT_VALUE

    def get_distance_to_target(
        self, flight_points: List[FlightPoint], target: FlightPoint
    ) -> float:
        current = flight_points[-1]
        return target.ground_distance - current.ground_distance


@dataclass
class OptimalCruiseSegment(CruiseSegment, mission_file_keyword="optimal_cruise"):
    """
    Class for computing cruise flight segment at maximum lift/drag ratio.

    Altitude is set **at every point** to get the optimum CL according to current mass.
    Target is a specified ground_distance. The target definition indicates
    the ground_distance to be covered during the segment, independently of
    the initial value.
    Target should also specify a speed parameter set to "constant", among `mach`,
    `true_airspeed` and `equivalent_airspeed`. If not, Mach will be assumed constant.
    """

    def compute_from_start_to_target(self, start: FlightPoint, target: FlightPoint) -> pd.DataFrame:
        start.altitude = self._get_optimal_altitude(start.mass, start.mach)
        self.complete_flight_point(start)
        return super().compute_from_start_to_target(start, target)

    def _compute_next_altitude(self, next_point: FlightPoint, previous_point: FlightPoint):
        next_point.altitude = self._get_optimal_altitude(
            next_point.mass, previous_point.mach, altitude_guess=previous_point.altitude
        )


@dataclass
class ClimbAndCruiseSegment(CruiseSegment, mission_file_keyword="cruise"):
    """
    Class for computing cruise flight segment at constant altitude.

    Target is a specified ground_distance. The target definition indicates
    the ground_distance to be covered during the segment, independently of
    the initial value.
    Target should also specify a speed parameter set to "constant", among `mach`,
    `true_airspeed` and `equivalent_airspeed`. If not, Mach will be assumed constant.

    Target altitude can also be set to
    :attr:`~.altitude_change.AltitudeChangeSegment.OPTIMAL_FLIGHT_LEVEL`. In that case, the cruise
    will be preceded by a climb segment and :attr:`climb_segment` must be set at instantiation.

    (Target ground distance will be achieved by the sum of ground distances
    covered during climb and cruise)

    In this case, climb will be done up to the IFR Flight Level (as multiple of 100 feet) that
    ensures minimum mass decrease, while being at most equal to :attr:`maximum_flight_level`.
    """

    #: The AltitudeChangeSegment that can be used if a preliminary climb is needed (its target
    #: will be ignored).
    climb_segment: AltitudeChangeSegment = None

    #: The maximum allowed flight level (i.e. multiple of 100 feet).
    maximum_flight_level: float = 500.0

    def compute_from_start_to_target(self, start: FlightPoint, target: FlightPoint) -> pd.DataFrame:
        climb_segment = deepcopy(self.climb_segment)
        climb_segment.target = target

        cruise_segment = CruiseSegment(
            target=deepcopy(target),  # deepcopy needed because altitude will be modified.
            propulsion=self.propulsion,
            reference_area=self.reference_area,
            polar=self.polar,
            name=self.name,
            engine_setting=self.engine_setting,
        )

        if (
            self.target.altitude == AltitudeChangeSegment.OPTIMAL_FLIGHT_LEVEL
            and climb_segment is not None
        ):
            cruise_segment.target.altitude = None

            # Go to the next flight level, or keep altitude if already at a flight level
            cruise_altitude = get_closest_flight_level(start.altitude - 1.0e-3)
            results = self._climb_to_altitude_and_cruise(
                start, cruise_altitude, climb_segment, cruise_segment
            )
            mass_loss = start.mass - results.mass.iloc[-1]

            go_to_next_level = True

            while go_to_next_level:
                old_mass_loss = mass_loss
                cruise_altitude = get_closest_flight_level(cruise_altitude + 1.0e-3)
                if cruise_altitude > self.maximum_flight_level * 100.0 * foot:
                    break

                new_results = self._climb_to_altitude_and_cruise(
                    start, cruise_altitude, climb_segment, cruise_segment
                )
                mass_loss = start.mass - new_results.mass.iloc[-1]

                go_to_next_level = mass_loss < old_mass_loss
                if go_to_next_level:
                    results = new_results

        elif target.altitude is not None:
            results = self._climb_to_altitude_and_cruise(
                start, target.altitude, climb_segment, cruise_segment
            )
        else:
            results = super().compute_from_start_to_target(start, target)

        return results

    @staticmethod
    def _climb_to_altitude_and_cruise(
        start: FlightPoint,
        cruise_altitude: float,
        climb_segment: AltitudeChangeSegment,
        cruise_segment: CruiseSegment,
    ):
        """
        Climbs up to cruise_altitude and cruise, while ensuring final ground_distance is
        equal to self.target.ground_distance.

        :param start:
        :param cruise_altitude:
        :param climb_segment:
        :param cruise_segment:
        :return:
        """
        climb_segment.target = FlightPoint(
            altitude=cruise_altitude,
            mach=cruise_segment.target.mach,
            true_airspeed=cruise_segment.target.true_airspeed,
            equivalent_airspeed=cruise_segment.target.equivalent_airspeed,
        )
        climb_points = climb_segment.compute_from(start)

        cruise_start = FlightPoint.create(climb_points.iloc[-1])
        cruise_points = cruise_segment.compute_from(cruise_start)

        return pd.concat([climb_points, cruise_points]).reset_index(drop=True)


@dataclass
class BreguetCruiseSegment(
    CruiseSegment,
    mission_file_keyword="breguet",
    attribute_units=dict(climb_and_descent_distance="m"),
):
    """
    Class for computing cruise flight segment at constant altitude using Breguet-Leduc formula.

    As formula relies on SFC, the :attr:`propulsion` model must be able to fill FlightPoint.sfc
    when FlightPoint.thrust is provided.
    """

    #: if True, max lift/drag ratio will be used instead of the one computed with polar using
    #: CL deduced from mass and altitude.
    #: In this case, reference_area parameter will be unused
    use_max_lift_drag_ratio: bool = False

    #: The reference area, in m**2. Used only if use_max_lift_drag_ratio is False.
    reference_area: float = 1.0

    def compute_from_start_to_target(self, start: FlightPoint, target: FlightPoint) -> pd.DataFrame:
        cruise_mass_ratio = self._compute_cruise_mass_ratio(
            start, target.ground_distance - start.ground_distance
        )

        end = deepcopy(start)
        end.mass = start.mass * cruise_mass_ratio
        end.ground_distance = target.ground_distance
        end.time = start.time + (end.ground_distance - start.ground_distance) / end.true_airspeed
        end.name = self.name
        self.complete_flight_point(end)

        return pd.DataFrame([start, end])

    def _compute_cruise_mass_ratio(self, start: FlightPoint, cruise_distance):
        """
        Computes mass ratio between end and start of cruise

        :param start: the initial flight point, defined for `CL`, `CD`, `mass` and`true_airspeed`
        :param cruise_distance: cruise distance in meters
        :return: (mass at end of cruise) / (mass at start of cruise)
        """

        if self.use_max_lift_drag_ratio:
            lift_drag_ratio = self.polar.optimal_cl / self.polar.cd(self.polar.optimal_cl)
        else:
            lift_drag_ratio = start.CL / start.CD
        start.thrust = start.mass / lift_drag_ratio * g
        self.propulsion.compute_flight_points(start)

        range_factor = start.true_airspeed * lift_drag_ratio / g / start.sfc
        return 1.0 / np.exp(cruise_distance / range_factor)



import numpy as np
import scipy.constants as sc

import openmdao.api as om


class ComputeFuelMass(om.ExplicitComponent):
    """
    Computes the fuel consumed during the mission based on the Breguet's range equation
    """

    def setup(self):

        # Defining the input(s)

        self.add_input(name="owe", units="kg", val=np.nan)
        self.add_input(name="payload", units="kg", val=np.nan)
        self.add_input(name="mission_range", units="m", val=np.nan)
        self.add_input(name="tsfc", units="kg/N/s", val=np.nan)
        self.add_input(name="l_d_ratio", val=np.nan)
        self.add_input(name="cruise_speed", units="m/s", val=np.nan)

        # Defining the output(s)

        self.add_output(name="mission_fuel", units="kg")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):

        # Assigning the input to local variable for clarity
        owe = inputs["owe"]
        payload = inputs["payload"]
        mission_range = inputs["mission_range"]
        tsfc = inputs["tsfc"]
        l_d = inputs["l_d_ratio"]
        cruise_speed = inputs["cruise_speed"]

        # To simplify the computation, we will first start by computing the range parameter,
        # which correspond to the term inside the exponential in the original formula
        range_parameter = (mission_range * tsfc * sc.g) / (cruise_speed * l_d)

        # Let's now computed the fuel using Breguet's range equation rearranged
        mission_fuel = (owe + payload) * (np.exp(range_parameter) - 1.0)

        outputs["mission_fuel"] = mission_fuel



"""Dataclass utilities."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import asdict, dataclass

#: To be put as default value for dataclass fields that should not have a default value.
#: See :class:`BaseDataClass` for further information.
MANDATORY_FIELD = object()


@dataclass
class BaseDataClass:
    """
    This class is a workaround for the following dataclass problem:

        If a dataclass defines a field with a default value, inheritor classes will not
        be allowed to define fields without default value, because then the non-default fields
        will follow a default field, which is forbidden.

    The chosen solution (from https://stackoverflow.com/a/53085935/16488238) is to always define
    default values, but mandatory fields will have the :const:`MANDATORY_FIELD` object as default.

    After initialization, :meth:`__post_init__` will process fields and raise an error if
    a field has :const:`MANDATORY_FIELD` as value.
    """

    def __post_init__(self):
        for name, value in asdict(self).items():
            if value is MANDATORY_FIELD:
                raise TypeError(f"__init__ missing 1 required argument: '{name}'")



"""
API
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import os.path as pth
import shutil
import sys
import textwrap as tw
from collections import defaultdict
from collections.abc import Iterable
from time import time
from typing import Dict, IO, List, Union

import openmdao.api as om
import pandas as pd
from IPython import InteractiveShell
from IPython.display import HTML, clear_output, display
from tabulate import tabulate

import fastoad.openmdao.whatsopt
from fastoad._utils.files import make_parent_dir
from fastoad._utils.resource_management.copy import copy_resource, copy_resource_folder
from fastoad.cmd.exceptions import (
    FastNoAvailableNotebookError,
    FastPathExistsError,
)
from fastoad.gui import OptimizationViewer, VariableViewer
from fastoad.io import IVariableIOFormatter
from fastoad.io.configuration import FASTOADProblemConfigurator
from fastoad.io.variable_io import DataFile
from fastoad.io.xml import VariableLegacy1XmlFormatter
from fastoad.module_management._bundle_loader import BundleLoader
from fastoad.module_management._plugins import DistributionPluginDefinition, FastoadLoader
from fastoad.module_management.service_registry import RegisterOpenMDAOSystem, RegisterPropulsion
from fastoad.openmdao.problem import FASTOADProblem
from fastoad.openmdao.variables import VariableList

DEFAULT_WOP_URL = "https://ether.onera.fr/whatsopt"
_LOGGER = logging.getLogger(__name__)

# Used for test purposes only
_PROBLEM_CONFIGURATOR = None


def get_plugin_information(print_data=False) -> Dict[str, DistributionPluginDefinition]:
    """
    Provides information about available FAST-OAD plugins.

    :param print_data: if True, plugin data are displayed.
    :return: a dict with installed package names as keys and matching FAST-OAD
             plugin definitions as values.
    """
    plugin_definitions = FastoadLoader().distribution_plugin_definitions
    if print_data:
        table_dicts = [plugin_def.to_dict() for plugin_def in plugin_definitions.values()]

        if table_dicts:
            table = pd.DataFrame(table_dicts)
            if InteractiveShell.initialized():
                display(HTML(table.to_html(index=False)))
            else:
                print(table.to_markdown(index=False))
        else:
            print("No available FAST-OAD plugin.")

    return plugin_definitions


def generate_notebooks(
    destination_path: str,
    overwrite: bool = False,
    distribution_name=None,
):
    """
    Copies notebook folder(s) from available plugin(s).

    :param destination_path: the inner structure of the folders will depend on
                             the number of installed package and the number of
                             plugins they contain.
    :param overwrite: if True and `destination_path` exists, it will be removed before writing.
    :param distribution_name: the name of an installed package that provides notebooks
    """
    # Check available notebooks
    folder_info_list = FastoadLoader().get_notebook_folder_list(distribution_name)

    if len(folder_info_list) == 0:
        raise FastNoAvailableNotebookError(distribution_name)

    # Create and copy folder
    destination_path = pth.abspath(destination_path)
    if pth.exists(destination_path):
        if overwrite:
            shutil.rmtree(destination_path)
        else:
            raise FastPathExistsError(
                f"Notebook folder {destination_path} not written because it already exists. "
                "Use overwrite=True to bypass.",
                destination_path,
            )

    # Write notebooks

    # We organize folder_info_list as dict of dict to easily count how many distributions
    # and plugin by distributions we have, which info is used for building target paths.
    folder_info_dict = defaultdict(dict)
    for folder_info in folder_info_list:
        folder_info_dict[folder_info.dist_name][folder_info.plugin_name] = folder_info

    only_one_dist = len(folder_info_dict) == 1
    for dist_content in folder_info_dict.values():
        only_one_plugin = len(dist_content) == 1
        for folder_info in dist_content.values():
            target_path_elements = [destination_path]
            if not only_one_dist:
                target_path_elements.append(folder_info.dist_name)
            if not only_one_plugin:
                target_path_elements.append(folder_info.plugin_name)

            target_path = pth.join(*target_path_elements)
            os.makedirs(target_path)
            copy_resource_folder(folder_info.package_name, target_path)

    return destination_path


def generate_configuration_file(
    configuration_file_path: str,
    overwrite: bool = False,
    distribution_name=None,
    sample_file_name=None,
):
    """
    Copies a sample configuration file from an available plugin.

    :param configuration_file_path: the path of file to be written
    :param overwrite: if True, the file will be written, even if it already exists
    :param distribution_name: the name of the installed package that provides the sample
                             configuration file (can be omitted if only one plugin is available)
    :param sample_file_name: the name of the sample configuration file (can be omitted if
                             the plugin provides only one configuration file)
    :return: path of generated file
    :raise FastPathExistsError: if overwrite==False and configuration_file_path already exists
    """

    if distribution_name is None:
        # If no distribution is specified, but only one contains configuration files, no need to
        # specify the name
        count_plugin_with_conf_file = 0
        dist_with_conf_file = ""
        plugin_config_files = {
            name: [resource.name for resource in definition.get_configuration_file_list()]
            for name, definition in get_plugin_information().items()
        }
        for plugin_name, conf_files in plugin_config_files.items():
            # Plugin is retained if it contains conf files. Additionally, if sample_file_name
            # is provided, plugin is retained only if sample_file_name is in provided conf files.
            if len(conf_files) > 0 and (sample_file_name is None or sample_file_name in conf_files):
                count_plugin_with_conf_file += 1
                dist_with_conf_file = plugin_name
            if count_plugin_with_conf_file > 1:
                break

        # If only one plugin has been retained, it can be used as automatically selected plugin.
        if count_plugin_with_conf_file == 1:
            distribution_name = dist_with_conf_file

    dist_definition = FastoadLoader().get_distribution_plugin_definition(distribution_name)
    file_info = dist_definition.get_configuration_file_info(sample_file_name)

    # Check on file overwrite
    configuration_file_path = pth.abspath(configuration_file_path)
    if not overwrite and pth.exists(configuration_file_path):
        raise FastPathExistsError(
            f"Configuration file {configuration_file_path} not written because it already exists. "
            "Use overwrite=True to bypass.",
            configuration_file_path,
        )

    # Write file
    make_parent_dir(configuration_file_path)
    copy_resource(file_info.package_name, file_info.name, configuration_file_path)
    _LOGGER.info('Sample configuration written in "%s".', configuration_file_path)

    return configuration_file_path


def generate_inputs(
    configuration_file_path: str,
    source_path: str = None,
    source_path_schema="native",
    overwrite: bool = False,
) -> str:
    """
    Generates input file for the problem specified in configuration_file_path.

    :param configuration_file_path: where the path of input file to write is set
    :param source_path: path of file data will be taken from
    :param source_path_schema: set to 'legacy' if the source file come from legacy FAST
    :param overwrite: if True, file will be written even if one already exists
    :return: path of generated file
    :raise FastPathExistsError: if overwrite==False and configuration_file_path already exists
    """
    conf = FASTOADProblemConfigurator(configuration_file_path)
    conf._set_configuration_modifier(_PROBLEM_CONFIGURATOR)

    input_file_path = conf.input_file_path
    if not overwrite and pth.exists(conf.input_file_path):
        raise FastPathExistsError(
            f"Input file {input_file_path} not written because it already exists. "
            "Use overwrite=True to bypass.",
            input_file_path,
        )

    if source_path_schema == "legacy":
        conf.write_needed_inputs(source_path, VariableLegacy1XmlFormatter())
    else:
        conf.write_needed_inputs(source_path)

    _LOGGER.info("Problem inputs written in %s", input_file_path)
    return input_file_path


def list_variables(
    configuration_file_path: str,
    out: Union[IO, str] = None,
    overwrite: bool = False,
    force_text_output: bool = False,
    tablefmt: str = "grid",
):
    """
    Writes list of variables for the problem specified in configuration_file_path.

    List is generally written as text. It can be displayed as a scrollable table view if:
    - function is used in an interactive IPython shell
    - out == sys.stdout
    - force_text_output == False

    :param configuration_file_path:
    :param out: the output stream or a path for the output file (None means sys.stdout)
    :param overwrite: if True and out parameter is a file path, the file will be written even if one
                      already exists
    :param force_text_output: if True, list will be written as text, even if command is used in an
                              interactive IPython shell (Jupyter notebook). Has no effect in other
                              shells or if out parameter is not sys.stdout
    :param tablefmt: The formatting of the requested table. Options are the same as those available
                     to the tabulate package. See tabulate.tabulate_formats for a complete list.
                     If "var_desc" the file will use the variable_descriptions.txt format.
    :return: path of generated file, or None if no file was generated.
    :raise FastPathExistsError: if `overwrite==False` and `out` is a file path and the file exists
    """
    if out is None:
        out = sys.stdout

    conf = FASTOADProblemConfigurator(configuration_file_path)
    conf._set_configuration_modifier(_PROBLEM_CONFIGURATOR)
    problem = conf.get_problem()
    problem.setup()

    # Extracting inputs and outputs
    variables = VariableList.from_problem(problem)
    variables.sort(key=lambda var: var.name)
    input_variables = VariableList([var for var in variables if var.is_input])
    output_variables = VariableList([var for var in variables if not var.is_input])

    for var in input_variables:
        var.metadata["I/O"] = "IN"
    for var in output_variables:
        var.metadata["I/O"] = "OUT"

    variables_df = (
        (input_variables + output_variables)
        .to_dataframe()[["name", "I/O", "desc"]]
        .rename(columns={"name": "NAME", "desc": "DESCRIPTION"})
    )

    if isinstance(out, str):
        out = pth.abspath(out)
        if not overwrite and pth.exists(out):
            raise FastPathExistsError(
                f"File {out} not written because it already exists. "
                "Use overwrite=True to bypass.",
                out,
            )
        make_parent_dir(out)
        out_file = open(out, "w", encoding="utf-8")
    else:
        if out == sys.stdout and InteractiveShell.initialized() and not force_text_output:
            display(HTML(variables_df.to_html(index=False)))
            return None

        # Here we continue with text output
        out_file = out

    if tablefmt == "var_desc":
        content = _generate_var_desc_format(variables_df)
    else:
        content = _generate_table_format(variables_df, tablefmt=tablefmt)

    out_file.write(content)

    if isinstance(out, str):
        out_file.close()
        _LOGGER.info("Output list written in %s", out)
        return out

    return None


def _generate_var_desc_format(variables_df):
    content = variables_df.to_csv(
        columns=["NAME", "DESCRIPTION"],
        sep="|",
        index=False,
        header=False,
        line_terminator="\n",
    ).replace("|", " || ")
    return content


def _generate_table_format(variables_df, tablefmt="grid"):
    # Break descriptions that are too long
    variables_df["DESCRIPTION"] = variables_df["DESCRIPTION"].apply(
        lambda s: "\n".join(tw.wrap(s, 100))
    )
    content = tabulate(
        variables_df, headers=variables_df.columns, showindex=False, tablefmt=tablefmt
    )
    return content + "\n"


def list_modules(
    source_path: Union[List[str], str] = None,
    out: Union[IO, str] = None,
    overwrite: bool = False,
    verbose: bool = False,
    force_text_output: bool = False,
):
    """
    Writes list of available systems.
    If source_path is given and if it defines paths where there are registered systems,
    they will be listed too.

    :param source_path: either a configuration file path, folder path, or list of folder path
    :param out: the output stream or a path for the output file (None means sys.stdout)
    :param overwrite: if True and out is a file path, the file will be written even if one already
                      exists
    :param verbose: if True, shows detailed information for each system
                    if False, shows only identifier and path of each system
    :param force_text_output: if True, list will be written as text, even if command is used in an
                              interactive IPython shell (Jupyter notebook). Has no effect in other
                              shells or if out parameter is not sys.stdout
    :return: path of generated file, or None if no file was generated.
    :raise FastPathExistsError: if `overwrite==False` and `out` is a file path and the file exists
    """
    if out is None:
        out = sys.stdout

    if isinstance(source_path, str):
        if pth.isfile(source_path):
            conf = FASTOADProblemConfigurator(source_path)
            conf._set_configuration_modifier(_PROBLEM_CONFIGURATOR)
            # As the problem has been configured,
            # BundleLoader now knows additional registered systems
        elif pth.isdir(source_path):
            RegisterOpenMDAOSystem.explore_folder(source_path)
        else:
            raise FileNotFoundError(f"Could not find {source_path}")
    elif isinstance(source_path, Iterable):
        for folder_path in source_path:
            if not pth.isdir(folder_path):
                _LOGGER.warning("SKIPPED %s: folder does not exist.", folder_path)
            else:
                RegisterOpenMDAOSystem.explore_folder(folder_path)
    elif source_path is not None:
        raise RuntimeError("Unexpected type for source_path")

    if verbose:
        cell_list = _get_detailed_system_list()
    else:
        cell_list = _get_simple_system_list()

    if isinstance(out, str):
        out = pth.abspath(out)
        if not overwrite and pth.exists(out):
            raise FastPathExistsError(
                f"File {out} not written because it already exists. "
                "Use overwrite=True to bypass.",
                out,
            )

        make_parent_dir(out)
        out_file = open(out, "w", encoding="utf-8")
    else:
        if (
            out == sys.stdout
            and InteractiveShell.initialized()
            and not force_text_output
            and not verbose
        ):
            display(HTML(tabulate(cell_list, tablefmt="html")))
            return None

        out_file = out

    out_file.write(tabulate(cell_list, tablefmt="grid"))
    out_file.write("\n")

    if isinstance(out, str):
        out_file.close()
        _LOGGER.info("System list written in %s", out)
        return out

    return None


def _get_simple_system_list():
    cell_list = [["   AVAILABLE MODULE IDENTIFIERS", "MODULE PATH"]]
    for identifier in sorted(RegisterOpenMDAOSystem.get_provider_ids()):
        path = BundleLoader().get_factory_path(identifier)
        cell_list.append([identifier, path])

    cell_list.append(["   AVAILABLE PROPULSION WRAPPER IDENTIFIERS", "MODULE PATH"])
    for identifier in sorted(RegisterPropulsion.get_provider_ids()):
        path = BundleLoader().get_factory_path(identifier)
        cell_list.append([identifier, path])

    return cell_list


def _get_detailed_system_list():
    cell_list = [["AVAILABLE MODULE IDENTIFIERS\n============================"]]
    for identifier in sorted(RegisterOpenMDAOSystem.get_provider_ids()):
        path = BundleLoader().get_factory_path(identifier)
        domain = RegisterOpenMDAOSystem.get_provider_domain(identifier)
        description = RegisterOpenMDAOSystem.get_provider_description(identifier)
        if description is None:
            description = ""

        # We remove OpenMDAO's native options from the description
        component = RegisterOpenMDAOSystem.get_system(identifier)
        component.options.undeclare("assembled_jac_type")
        component.options.undeclare("distributed")

        cell_content = (
            f"IDENTIFIER:   {identifier}\n"
            f"PATH:         {path}\n"
            f"DOMAIN:       {domain.value}\n"
            f"DESCRIPTION:  {tw.indent(tw.dedent(description), '    ')}\n"
        )
        if len(list(component.options.items())) > 0:
            cell_content += component.options.to_table(fmt="grid") + "\n"

        cell_list.append([cell_content])
    cell_list.append(
        ["AVAILABLE PROPULSION WRAPPER IDENTIFIERS\n========================================"]
    )
    for identifier in sorted(RegisterPropulsion.get_provider_ids()):
        path = BundleLoader().get_factory_path(identifier)
        description = RegisterPropulsion.get_provider_description(identifier)
        if description is None:
            description = ""

        cell_content = (
            f"IDENTIFIER:   {identifier}\n"
            f"PATH:         {path}\n"
            f"DESCRIPTION:  {tw.indent(tw.dedent(description), '    ')}\n"
        )
        cell_list.append([cell_content])
    return cell_list


def write_n2(configuration_file_path: str, n2_file_path: str = None, overwrite: bool = False):
    """
    Write the N2 diagram of the problem in file n2.html

    :param configuration_file_path:
    :param n2_file_path: if None, will default to `n2.html`
    :param overwrite:
    :return: path of generated file.
    :raise FastPathExistsError: if overwrite==False and n2_file_path already exists
    """

    if not n2_file_path:
        n2_file_path = "n2.html"
    n2_file_path = pth.abspath(n2_file_path)

    if not overwrite and pth.exists(n2_file_path):
        raise FastPathExistsError(
            f"N2-diagram file {n2_file_path} not written because it already exists. "
            "Use overwrite=True to bypass.",
            n2_file_path,
        )

    make_parent_dir(n2_file_path)
    conf = FASTOADProblemConfigurator(configuration_file_path)
    conf._set_configuration_modifier(_PROBLEM_CONFIGURATOR)
    problem = conf.get_problem()
    problem.setup()
    problem.final_setup()

    om.n2(problem, outfile=n2_file_path, show_browser=False)
    if InteractiveShell.initialized():
        clear_output()
    _LOGGER.info("N2 diagram written in %s", pth.abspath(n2_file_path))
    return n2_file_path


def write_xdsm(
    configuration_file_path: str,
    xdsm_file_path: str = None,
    overwrite: bool = False,
    depth: int = 2,
    wop_server_url: str = None,
    dry_run: bool = False,
):
    """

    :param configuration_file_path:
    :param xdsm_file_path: the path for HTML file to be written (will overwrite if needed)
    :param overwrite: if False, will raise an error if file already exists.
    :param depth: the depth analysis for WhatsOpt
    :param wop_server_url: URL of WhatsOpt server (if None, ether.onera.fr/whatsopt will be used)
    :param dry_run: if True, will run wop without sending any request to the server. Generated
                    XDSM will be empty. (for test purpose only)
    :return: path of generated file.
    :raise FastPathExistsError: if overwrite==False and xdsm_file_path already exists
    """
    if not xdsm_file_path:
        xdsm_file_path = pth.join(pth.dirname(configuration_file_path), "xdsm.html")
    xdsm_file_path = pth.abspath(xdsm_file_path)

    if not overwrite and pth.exists(xdsm_file_path):
        raise FastPathExistsError(
            "XDSM-diagram file %s not written because it already exists. "
            "Use overwrite=True to bypass." % xdsm_file_path,
            xdsm_file_path,
        )

    make_parent_dir(xdsm_file_path)

    conf = FASTOADProblemConfigurator(configuration_file_path)
    conf._set_configuration_modifier(_PROBLEM_CONFIGURATOR)
    problem = conf.get_problem()
    problem.setup()
    problem.final_setup()

    fastoad.openmdao.whatsopt.write_xdsm(problem, xdsm_file_path, depth, wop_server_url, dry_run)
    return xdsm_file_path


def _run_problem(
    configuration_file_path: str,
    overwrite: bool = False,
    mode="run_model",
    auto_scaling: bool = False,
) -> FASTOADProblem:
    """
    Runs problem according to provided file

    :param configuration_file_path: problem definition
    :param overwrite: if True, output file will be overwritten
    :param mode: 'run_model' or 'run_driver'
    :param auto_scaling: if True, automatic scaling is performed for design variables and
                         constraints
    :return: the OpenMDAO problem after run
    :raise FastPathExistsError: if overwrite==False and output data file of problem already exists
    """

    conf = FASTOADProblemConfigurator(configuration_file_path)
    conf._set_configuration_modifier(_PROBLEM_CONFIGURATOR)
    problem = conf.get_problem(read_inputs=True, auto_scaling=auto_scaling)

    outputs_path = pth.normpath(problem.output_file_path)
    if not overwrite and pth.exists(outputs_path):
        raise FastPathExistsError(
            f"Problem not run because output file {outputs_path} already exists. "
            "Use overwrite=True to bypass.",
            outputs_path,
        )

    problem.setup()

    start_time = time()
    if mode == "run_model":
        problem.run_model()
        problem.optim_failed = False  # Actually, we don't know
    else:
        problem.optim_failed = problem.run_driver()
    end_time = time()
    computation_time = round(end_time - start_time, 2)

    problem.write_outputs()
    if problem.optim_failed:
        _LOGGER.error("Optimization failed after %s seconds", computation_time)
    else:
        _LOGGER.info("Computation finished after %s seconds", computation_time)

    _LOGGER.info("Problem outputs written in %s", outputs_path)

    return problem


def evaluate_problem(configuration_file_path: str, overwrite: bool = False) -> FASTOADProblem:
    """
    Runs model according to provided problem file

    :param configuration_file_path: problem definition
    :param overwrite: if True, output file will be overwritten
    :return: the OpenMDAO problem after run
    :raise FastPathExistsError: if overwrite==False and output data file of problem already exists
    """
    return _run_problem(configuration_file_path, overwrite, "run_model")


def optimize_problem(
    configuration_file_path: str, overwrite: bool = False, auto_scaling: bool = False
) -> FASTOADProblem:
    """
    Runs driver according to provided problem file

    :param configuration_file_path: problem definition
    :param overwrite: if True, output file will be overwritten
    :param auto_scaling: if True, automatic scaling is performed for design variables and
                         constraints
    :return: the OpenMDAO problem after run
    :raise FastPathExistsError: if overwrite==False and output data file of problem already exists
    """
    return _run_problem(configuration_file_path, overwrite, "run_driver", auto_scaling=auto_scaling)


def optimization_viewer(configuration_file_path: str):
    """
    Displays optimization information and enables its editing

    :param configuration_file_path: problem definition
    :return: display of the OptimizationViewer
    """
    conf = FASTOADProblemConfigurator(configuration_file_path)
    conf._set_configuration_modifier(_PROBLEM_CONFIGURATOR)
    viewer = OptimizationViewer()
    viewer.load(conf)

    return viewer.display()


def variable_viewer(file_path: str, file_formatter: IVariableIOFormatter = None, editable=True):
    """
    Displays a widget that enables to visualize variables information and edit their values.

    :param file_path: the path of file to interact with
    :param file_formatter: the formatter that defines file format. If not provided, default format
                           will be assumed.
    :param editable: if True, an editable table with variable filters will be displayed. If False,
                     the table will not be editable nor searchable, but can be stored in an HTML
                     file.
    :return: display handle of the VariableViewer
    """
    if editable:
        viewer = VariableViewer()
        viewer.load(file_path, file_formatter)

        handle = viewer.display()
    else:
        variables = DataFile(file_path, formatter=file_formatter)
        variables.sort(key=lambda var: var.name)

        table = variables.to_dataframe()[["name", "val", "units", "is_input", "desc"]].rename(
            columns={"name": "Name", "val": "Value", "units": "Unit", "desc": "Description"}
        )
        table["I/O"] = "OUT"
        table["I/O"].loc[table["is_input"]] = "IN"
        del table["is_input"]
        table.set_index("Name", drop=True, inplace=True)

        handle = display(HTML(table.to_html()))
    return handle



"""Utility functions for CLI interface."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Callable

import click

from fastoad.cmd.exceptions import FastPathExistsError


def overwrite_option(func):
    """
    Decorator for adding the option for overwriting existing file.

    Use `force` as argument of the function.
    """
    return click.option(
        "-f",
        "--force",
        is_flag=True,
        help="Do not ask before overwriting files.",
    )(func)


def out_file_option(func):
    """
    Decorator for writing command output in a file.

    Use `out_file` and `force` as argument of the function.
    """
    return click.option(
        "-o",
        "--out_file",
        help="If provided, command output will be written in indicated file instead of "
        "being printed in terminal.",
    )(overwrite_option(func))


def manage_overwrite(func: Callable, filename_func: Callable = None, **kwargs):
    """
    Runs `func`, that is expected to write a file, with provided keyword arguments `args`.

    If the run throws FastPathExistsError, a question is displayed and user is
    asked for a yes/no answer. If `yes` is given, arg["overwrite"] is set to True
    and `func` is run again.

    :param func: callable that will do the operation and is expected to return
                 the path of written element.
    :param filename_func: a function that provides the name of written file, given the
                          value returned by func
    :param kwargs: keyword arguments for func
    :return: True if the file has been written,
    """
    written = False
    try:
        written = _run_func(func, filename_func, **kwargs)

    except FastPathExistsError as exc:
        if click.confirm(f'"{exc.args[1]}" already exists. Do you want to overwrite it?'):
            kwargs["overwrite"] = True
            written = _run_func(func, filename_func, **kwargs)
        else:
            click.echo("Operation cancelled.")

    return written


def _run_func(func: Callable, filename_func: Callable = None, **kwargs):
    result = func(**kwargs)
    if result:
        if filename_func:
            result = filename_func(result)
        click.echo(f'"{result}" has been written.')
        return True

    return False



"""WhatsOpt-related operations."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2021 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import openmdao.api as om
import whatsopt.whatsopt_client as wop

from fastoad._utils.files import make_parent_dir
from fastoad import __version__ as fastoad_version

_LOGGER = logging.getLogger(__name__)  # Logger for this module


def write_xdsm(
    problem: om.Problem,
    xdsm_file_path: str = None,
    depth: int = 2,
    wop_server_url: str = None,
    dry_run: bool = False,
):
    """
    Makes WhatsOpt generate a XDSM in HTML file.

    :param problem: a Problem instance. final_setup() must have been run.
    :param xdsm_file_path: the path for HTML file to be written (will overwrite if needed)
    :param depth: the depth analysis for WhatsOpt
    :param wop_server_url: URL of WhatsOpt server (if None, ether.onera.fr/whatsopt will be used)
    :param dry_run: if True, will run wop without sending any request to the server. Generated
                    XDSM will be empty. (for test purpose only)
    """

    make_parent_dir(xdsm_file_path)

    if not wop_server_url:
        wop_server_url = wop.EXTRANET_SERVER_URL

    wop_session = wop.WhatsOpt(url=wop_server_url)

    xdsm = wop_session.push_mda(
        problem, {"--xdsm": True, "--name": None, "--depth": depth, "--dry-run": dry_run}
    )
    wop.generate_xdsm_html(f"FAST-OAD {fastoad_version}", xdsm=xdsm, outfilename=xdsm_file_path)



"""
Utilities for mission computation.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy.constants import foot

FLIGHT_LEVEL = 100 * foot


def get_closest_flight_level(altitude, base_level=0, level_step=10, up_direction=True):
    """
    Computes the altitude (in meters) of a flight level close to provided altitude.

    Flight levels are multiples of 100 feet.

    see examples below::

        >>> # Getting the IFR flight level immediately above
        >>> get_closest_flight_level(4400. * foot)
        5000.0
        >>> # Getting the IFR flight level immediately below
        >>> get_closest_flight_level(4400. * foot, up_direction=False)
        4000.0
        >>> # Getting the next even IFR flight level
        >>> get_closest_flight_level(4400. * foot, level_step = 20)
        6000.0
        >>> # Getting the next odd IFR flight level
        >>> get_closest_flight_level(3100. * foot, base_level=10, level_step = 20)
        5000.0

    :param altitude: in meters
    :param base_level: base flight level for computed steps
    :param level_step: number of flight level per step
    :param up_direction: True if next flight level is upper. False if lower
    :return: the altitude in meters of the asked flight level.
    """
    if up_direction:
        func = np.ceil
    else:
        func = np.floor

    base_altitude = FLIGHT_LEVEL * base_level
    return base_altitude + FLIGHT_LEVEL * level_step * func(
        (altitude - base_altitude) / FLIGHT_LEVEL / level_step
    )



"""Base classes for mission computation."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field
from typing import List, Optional

import pandas as pd

from fastoad.model_base import FlightPoint
from fastoad.model_base.datacls import BaseDataClass


@dataclass
class IFlightPart(ABC, BaseDataClass):
    """
    Base class for all flight parts.
    """

    name: str = ""
    target: FlightPoint = field(init=False, default=None)

    @abstractmethod
    def compute_from(self, start: FlightPoint) -> pd.DataFrame:
        """
        Computes a flight sequence from provided start point.

        :param start: the initial flight point, defined for `altitude`, `mass` and speed
                      (`true_airspeed`, `equivalent_airspeed` or `mach`). Can also be
                      defined for `time` and/or `ground_distance`.
        :return: a pandas DataFrame where column names match fields of
                 :class:`~fastoad.model_base.flight_point.FlightPoint`
        """


@dataclass
class FlightSequence(IFlightPart):
    """
    Defines and computes a flight sequence.
    """

    #: List of IFlightPart instances that should be run sequentially.
    flight_sequence: List[IFlightPart] = field(default_factory=list)

    #: Consumed mass between sequence start and target mass, if any defined
    consumed_mass_before_input_weight: float = field(default=0.0, init=False)

    def compute_from(self, start: FlightPoint) -> pd.DataFrame:
        parts = []
        part_start = deepcopy(start)
        part_start.scalarize()

        self.consumed_mass_before_input_weight = 0.0
        consumed_mass = 0.0
        for part in self.flight_sequence:
            # This check has to be done first because relative target parameters
            # will be made absolute during compute_from()
            part_has_target_mass = not (part.target.mass is None or part.target.is_relative("mass"))

            flight_points = part.compute_from(part_start)

            consumed_mass += flight_points.iloc[0].mass - flight_points.iloc[-1].mass

            if isinstance(part, FlightSequence):
                self.consumed_mass_before_input_weight += part.consumed_mass_before_input_weight

            # If a part has a target mass, computed mass of all previous part must be
            # offset so this target will be reached.
            # (mass consumption of previous parts is assumed independent of aircraft mass)
            if part_has_target_mass:
                mass_offset = flight_points.iloc[0].mass - part_start.mass
                for previous_flight_points in parts:
                    previous_flight_points.mass += mass_offset
                self.consumed_mass_before_input_weight = float(consumed_mass)

            if len(parts) > 0 and len(flight_points) > 1:
                # First point of the segment is omitted, as it is the last of previous segment.
                #
                # But sometimes (especially in the case of simplistic segments), the new first
                # point may contain more information than the previous last one. In such case,
                # it is interesting to complete the previous last one.
                last_flight_points = parts[-1]
                last_index = last_flight_points.index[-1]
                for name in flight_points.columns:
                    value = last_flight_points.loc[last_index, name]
                    if not value:
                        last_flight_points.loc[last_index, name] = flight_points.loc[0, name]

                parts.append(flight_points.iloc[1:])

            else:
                # But it is kept if the computed segment is the first one.
                parts.append(flight_points)

            part_start = FlightPoint.create(flight_points.iloc[-1])
            part_start.scalarize()

        if parts:
            return pd.concat(parts).reset_index(drop=True)

    @property
    def target(self) -> Optional[FlightPoint]:
        """Target of the last element of current sequence."""
        if len(self.flight_sequence) > 0:
            return self.flight_sequence[-1].target

        return None



"""
Defines the variable viewer for postprocessing
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2021 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, List, Set

import ipysheet as sh
import ipywidgets as widgets
import pandas as pd
from IPython.display import display

from fastoad.io import IVariableIOFormatter, VariableIO
from fastoad.openmdao.variables import VariableList

pd.set_option("display.max_rows", None)
INPUT = "IN"
OUTPUT = "OUT"
NA = "N/A"
TAG_ALL = "--ALL--"


class VariableViewer:
    """
    A class for interacting with FAST-OAD files.
    The file data is stored in a pandas DataFrame. The class built so that a modification
    of the DataFrame is instantly replicated on the file file.
    The interaction is achieved using a user interface built with widgets from ipywidgets and
    Sheets from ipysheet.

    A classical usage of this class will be::

        df = VariableViewer()  # instantiation of dataframe
        file = AbstractOMFileIO('problem_outputs.file') #  instantiation of file io
        df.load(file)  # load the file
        df.display()  # renders a ui for reading/modifying the file
    """

    # When getting a dataframe from a VariableList, the dictionary keys tell
    # what columns are kept and the values tell what name will be displayed.
    _DEFAULT_COLUMN_RENAMING = {
        "name": "Name",
        "val": "Value",
        "units": "Unit",
        "desc": "Description",
        "is_input": "is_input",
    }

    def __init__(self):

        #: The path of the data file that will be viewed/edited
        self.file = None

        #: The dataframe which is the mirror of self.file
        self.dataframe = pd.DataFrame()

        # The sheet which is the mirror of the dataframe
        self._sheet = None

        # The list of stored widgets
        self._filter_widgets = None

        # The ui containing save and load buttons
        self._save_load_buttons = None

        # The ui containing all the dropdown menus
        self._variable_selector = None

        # The ui containing all the input/output selector
        self._io_selector = None

        # The complete user interface
        self._ui = None

    def load(self, file_path: str, file_formatter: IVariableIOFormatter = None):
        """
        Loads the file and stores its data.

        :param file_path: the path of file to interact with
        :param file_formatter: the formatter that defines file format. If not
                               provided, default format will be assumed.
        """

        self.file = file_path
        self.load_variables(VariableIO(file_path, file_formatter).read())

    def save(self, file_path: str = None, file_formatter: IVariableIOFormatter = None):
        """
        Save the dataframe to the file.

        :param file_path: the path of file to save. If not given, the initially
                          read file will be overwritten.
        :param file_formatter: the formatter that defines file format. If not
                               provided, default format will be assumed.
        """
        if file_path is None:
            file_path = self.file

        variables = self.get_variables()

        VariableIO(file_path, file_formatter).write(variables)

    def display(self):
        """
        Displays the datasheet
        :return display of the user interface:
        """
        self._create_save_load_buttons()
        self._create_io_selector()
        return self._render_sheet()

    def load_variables(self, variables: VariableList, attribute_to_column: Dict[str, str] = None):
        """
        Loads provided variable list and replace current data set.

        :param variables: the variables to load
        :param attribute_to_column: dictionary keys tell what variable attributes are kept
                                    and the values tell what name will be displayed. If
                                    not provided, default translation will apply.
        """

        if not attribute_to_column:
            attribute_to_column = self._DEFAULT_COLUMN_RENAMING

        self.dataframe = (
            variables.to_dataframe()
            .rename(columns=attribute_to_column)[attribute_to_column.values()]
            .reset_index(drop=True)
        )

        self.dataframe["I/O"] = NA
        self.dataframe.loc[self.dataframe["is_input"] == 1, "I/O"] = INPUT
        self.dataframe.loc[self.dataframe["is_input"] == 0, "I/O"] = OUTPUT
        self.dataframe.drop(columns=["is_input"], inplace=True)

    def get_variables(self, column_to_attribute: Dict[str, str] = None) -> VariableList:
        """

        :param column_to_attribute: dictionary keys tell what columns are kept and the values
                                    tell what variable attribute it corresponds to. If not
                                    provided, default translation will apply.
        :return: a variable list from current data set
        """
        if not column_to_attribute:
            column_to_attribute = {
                value: key for key, value in self._DEFAULT_COLUMN_RENAMING.items()
            }

        dataframe = self.dataframe.copy()

        dataframe["is_input"] = None
        dataframe.loc[dataframe["I/O"] == INPUT, "is_input"] = True
        dataframe.loc[dataframe["I/O"] == OUTPUT, "is_input"] = False
        dataframe.drop(columns=["I/O"], inplace=True)
        variables = VariableList.from_dataframe(
            dataframe[column_to_attribute.keys()].rename(columns=column_to_attribute)
        )
        return variables

    # pylint: disable=invalid-name # df is a common naming for dataframes
    @staticmethod
    def _df_to_sheet(df: pd.DataFrame) -> sh.Sheet:
        """
        Transforms a pandas DataFrame into a ipysheet Sheet.
        The cells are set to read only except for the values.

        :param df: the pandas DataFrame to be converted
        :return the equivalent ipysheet Sheet
        """
        if not df.empty:
            sheet = sh.from_dataframe(df)
            column = df.columns.get_loc("Value")

            for cell in sheet.cells:
                if column not in (cell.column_start, cell.column_end):
                    cell.read_only = True
                else:
                    cell.type = "numeric"
                    # TODO: make the number of decimals depend on the module ?
                    # or chosen in the ui by the user
                    cell.numeric_format = "0.000"

            # Name, Value, Unit, Description
            sheet.column_width = [150, 50, 20, 150]

        else:
            sheet = sh.sheet()
        return sheet

    @staticmethod
    def _sheet_to_df(sheet: sh.Sheet) -> pd.DataFrame:
        """
        Transforms a ipysheet Sheet into a pandas DataFrame.

        :param sheet: the ipysheet Sheet to be converted
        :return the equivalent pandas DataFrame
        """
        df = sh.to_dataframe(sheet)
        return df

    # pylint: disable=unused-argument  # args has to be there for observe() to work
    def _update_df(self, change=None):
        """
        Updates the stored DataFrame with respect to the actual values of the Sheet.
        Then updates the file with respect to the stored DataFrame.
        """
        df = self._sheet_to_df(self._sheet)
        for i in df.index:
            self.dataframe.loc[int(i), :] = df.loc[i, :].values

    def _render_sheet(self) -> display:
        """
        Renders an interactive pysheet with dropdown menus of the stored dataframe.

        :return display of the user interface
        """
        self._filter_widgets = []
        modules_item = [TAG_ALL] + sorted(self._find_submodules(self.dataframe))
        w = widgets.Dropdown(options=modules_item)
        self._filter_widgets.append(w)
        return self._render_ui()

    # pylint: disable=unused-argument  # args has to be there for observe() to work
    def _update_items(self, change=None):
        """
        Updates the filter_widgets with respect to higher level filter_widgets values.
        """
        # 20 will never be reached
        for i in range(20):
            if i == 0:
                self._filter_widgets[0].observe(self._update_items, "value")
                self._filter_widgets[0].observe(self._update_variable_selector, "value")
            elif i <= len(self._filter_widgets):
                modules = [item.value for item in self._filter_widgets[0:i]]
                var_name = ":".join(modules)
                modules_item = sorted(self._find_submodules(self.dataframe, modules))
                if modules_item:
                    # Check if the item exists already
                    if i == len(self._filter_widgets):
                        if len(modules_item) > 1:
                            modules_item.insert(0, TAG_ALL)
                        widget = widgets.Dropdown(options=modules_item)
                        widget.observe(self._update_items, "value")
                        widget.observe(self._update_variable_selector, "value")
                        self._filter_widgets.append(widget)
                    else:
                        if (TAG_ALL not in modules_item) and (
                            len(modules_item) > 1 or var_name in self.dataframe["Name"].values
                        ):
                            modules_item.insert(0, TAG_ALL)
                        self._filter_widgets[i].options = modules_item
                else:
                    if i < len(self._filter_widgets):
                        self._filter_widgets.pop(i)
            else:
                break

    # pylint: disable=unused-argument  # args has to be there for observe() to work
    def _update_variable_selector(self, change=None):
        """
        Updates the variable selector with respect to the
        actual filter_widgets stored.
        """
        items_box = widgets.HBox(self._filter_widgets)
        items_box = widgets.VBox([widgets.Label(value="Variable name"), items_box])
        self._variable_selector = items_box

    def _create_io_selector(self):
        """
        The dropdown menu enables to selector only inputs, only outputs or all variables.
        """
        io_selector = widgets.Dropdown(
            options=[TAG_ALL, "Inputs", "Outputs"], layout=widgets.Layout(width="auto")
        )
        io_selector.observe(self._render_ui, "value")

        self._io_selector = widgets.VBox([widgets.Label(value="I/O"), io_selector])

    def _create_save_load_buttons(self):
        """
        The save button saves the present state of the dataframe to the xml.
        The load button loads the xml and replaces actual the dataframe.
        """

        save_button = widgets.Button(
            description="Save",
            disabled=False,
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Save to the file",
            icon="save",
        )

        def on_save_button_clicked(b):
            self.save()

        save_button.on_click(on_save_button_clicked)

        load_button = widgets.Button(
            description="Load",
            disabled=False,
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Load the file",
            icon="upload",
        )

        def on_load_button_clicked(b):
            self.load(self.file)
            self._render_sheet()

        load_button.on_click(on_load_button_clicked)

        items_box = widgets.HBox([load_button, save_button])

        self._save_load_buttons = items_box

    def _update_sheet(self):
        """
        Updates the sheet after filtering the dataframe with respect to
        the actual values of the variable dropdown menus.
        """
        modules = [item.value for item in self._filter_widgets]
        io_value = self._io_selector.children[1].value
        if io_value == "Inputs":
            var_io_type = INPUT
        elif io_value == "Outputs":
            var_io_type = OUTPUT
        else:
            var_io_type = TAG_ALL

        filtered_var = self._filter_variables(self.dataframe, modules, var_io_type=var_io_type)

        self._sheet = self._df_to_sheet(filtered_var)

        for cell in self._sheet.cells:
            cell.observe(self._update_df, "value")

    # pylint: disable=unused-argument  # args has to be there for observe() to work
    def _render_ui(self, change=None) -> display:
        """
        Renders the dropdown menus for the variable selector and the corresponding
        ipysheet Sheet containing the variable info.

        :return the display object
        """
        self._update_items()
        self._update_variable_selector()
        self._update_sheet()
        for item in self._filter_widgets:
            item.observe(self._render_ui, "value")
        self._sheet.layout.height = "400px"
        selectors = widgets.HBox([self._io_selector, self._variable_selector])

        if self._ui is not None:
            self._ui.children = [self._save_load_buttons, selectors, self._sheet]
            return None

        self._ui = widgets.VBox([self._save_load_buttons, selectors, self._sheet])
        return display(self._ui)

    @staticmethod
    def _find_submodules(df: pd.DataFrame, modules: List[str] = None) -> Set[str]:
        """
        Search for submodules at root or provided modules.

        To find the submodules the method analyzes the name of the variables.
        If the kwarg 'modules' is not None, the submodules search will be applied to
        the variables that are part of these modules.

        :param df: the pandas dataframe containing the variables
        :param modules: the list of modules to which the variables belong
        :return the submodules list
        """
        var_names = df.filter(items=["Name"])

        if not modules:
            modules = []

        def get_next_module(path):
            submodules = path.split(":")
            if len(modules) >= len(submodules) or submodules[: len(modules)] != modules:
                return ""

            return submodules[len(modules)]

        submodules = var_names.applymap(get_next_module)
        submodules = submodules[submodules.Name != ""]

        return set(submodules["Name"].tolist())

    @staticmethod
    def _filter_variables(
        df: pd.DataFrame, modules: List[str], var_io_type: str = None
    ) -> pd.DataFrame:
        """
        Returns a filtered dataframe with respect to a set of modules and variable type.

        The variables kept must be part of the modules list provided and the variable type
        'IN' or 'OUT'(if provided).

        :param df: the pandas dataframe containing the variables
        :param modules: the list of modules to which the variables belong
        :param var_io_type: the type of variables to keep
        :return the filtered dataframe
        """
        if var_io_type is None:
            var_io_type = TAG_ALL
        path = ""
        for _ in modules:
            if modules[-1] == TAG_ALL:
                path = ":".join(modules[:-1])
            else:
                path = ":".join(modules)
        path_filter = df.Name.str.startswith(path)
        io_filter = [True] * len(df) if var_io_type == TAG_ALL else df["I/O"] == var_io_type

        filtered_df = df[path_filter & io_filter]

        return filtered_df



"""For checking validity domain of OpenMDAO variables."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2021 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import inspect
import logging
import uuid
from collections import namedtuple
from dataclasses import dataclass
from enum import IntEnum
from typing import Dict, List
from uuid import UUID

import numpy as np
import openmdao.api as om
from openmdao.utils.units import convert_units

from fastoad.openmdao.variables import VariableList

CheckRecord = namedtuple(
    "CheckRecord",
    [
        "variable_name",
        "status",
        "limit_value",
        "limit_units",
        "value",
        "value_units",
        "source_file",
        "logger_name",
    ],
)
"""
A namedtuple that contains result of one variable check
"""


class ValidityStatus(IntEnum):
    """
    Simple enumeration for validity status.
    """

    OK = 0
    TOO_LOW = -1
    TOO_HIGH = 1


class ValidityDomainChecker:
    """
    Decorator class that checks variable values against limit bounds

    This class aims at producing a status of out of limits variables at the end
    of an OpenMDAO computation.

    The point is to allow to define limit bounds when defining an OpenMDAO
    system, but to make the check on the OpenMDAO problem after the run.

    When defining an OpenMDAO system, use this class as Python decorator to
    define validity domains:

    .. code-block::

        @ValidityDomainChecker
        class MyComponent(om.explicitComponent):
            ...

    The above code will check values against lower and upper bounds that have
    been defined when adding OpenMDAO outputs.

    Next code shows how to define lower and upper bounds, for inputs and/or
    outputs.

    .. code-block::

        @ValidityDomainChecker(
            {
                "a:variable:with:two:bounds": (-10.0, 1.0),
                "a:variable:with:lower:bound:only": (0.0, None),
                "a:variable:with:upper:bound:only": (None, 4.2),
            },
        )
        class MyComponent(om.explicitComponent):
            ...

    The defined domain limits supersedes lower and upper bounds from
    OpenMDAO output definitions, but only in the frame of ValidityDomainChecker.
    In any case, OpenMDAO process is not affected by usage of ValidityDomainChecker.

    Validity status can be obtained through log messages from Python logging module
    after problem has been run with:

    .. code-block::

        ...
        problem.run_model()
        ValidityDomainChecker.check_problem_variables(problem)

    **Warnings**:
    - Units of limit values defined in ValidityDomainChecker are assumed to be the
      same as in add_input() and add_output() statements of decorated class
    - Validity check currently only applies to scalar values
    """

    _limit_definitions: Dict[UUID, "_LimitDefinitions"] = {}

    def __init__(self, limits: Dict[str, tuple] = None, logger_name: str = None):
        """
        :param limits: a dictionary where keys are variable names and values are two-values tuples
                   that give lower and upper bound. One bound can be set to None.
        :param logger_name: The named of the logger that will be used. If not provided, name of
                        current module (i.e. "__name__"") will be used.
        """
        self._uuid = uuid.uuid4()

        # use given logger name for now. If it is None, it will be replaced by
        # calling module name in __call__ (because I could not find a way to get
        # this module name here with inspect, though it is very easy when we
        # have the decorated class)
        limit_definitions = _LimitDefinitions(self._get_caller_filename(), logger_name)
        for var_name, bounds in limits.items():
            lower = bounds[0] if bounds[0] is not None else -np.inf
            upper = bounds[1] if bounds[1] is not None else np.inf
            limit_definitions[var_name] = _LimitDefinition(lower, upper)

        self._limit_definitions[self._uuid] = limit_definitions

    def __call__(self, om_class: type):
        # Update logger name if needed: if it was not given, module name of
        # decorated class is used.
        if not self._limit_definitions[self._uuid].logger_name:
            self._limit_definitions[self._uuid].logger_name = om_class.__module__

        # We add to the OpenMDAO class a reference to self._limit_definitions[self._uuid]
        # to be able to update it with OpenMDAO declarations after problem setup.
        # See _update_problem_limit_definitions()
        om_class._fastoad_limit_definitions = self._limit_definitions[
            self._uuid
        ]  # pylint: disable=protected-access # We create it
        return om_class

    @classmethod
    def check_problem_variables(cls, problem: om.Problem) -> List[CheckRecord]:
        """
        Checks variable values in provided problem.

        Logs warnings for each variable that is out of registered limits.

        problem.setup() must have been run.

        :param problem:
        :return: the list of checks
        """

        cls._update_problem_limit_definitions(problem)

        variables = VariableList.from_problem(problem)
        records = cls.check_variables(variables)
        cls.log_records(records)
        return records

    @classmethod
    def check_variables(cls, variables: VariableList) -> List[CheckRecord]:
        """
        Check values of provided variables against registered limits.

        :param variables:
        :return: the list of checks
        """
        records: List[CheckRecord] = []

        for var in variables:
            for limit_definitions in cls._limit_definitions.values():
                if var.name in limit_definitions:
                    limit_def = limit_definitions[var.name]
                    value = convert_units(var.value, var.units, limit_def.units)
                    if value < limit_def.lower:
                        status = ValidityStatus.TOO_LOW
                        limit = limit_def.lower
                    elif value > limit_def.upper:
                        status = ValidityStatus.TOO_HIGH
                        limit = limit_def.upper
                    else:
                        status = ValidityStatus.OK
                        limit = None

                    records.append(
                        CheckRecord(
                            var.name,
                            status,
                            limit,
                            limit_def.units,
                            var.value,
                            var.units,
                            limit_definitions.source_file,
                            limit_definitions.logger_name,
                        )
                    )
        return records

    @staticmethod
    def log_records(records: List[CheckRecord]):
        """
        Logs warnings through Python logging module for each CheckRecord in provided list if
        it is not OK.

        :param records:
        :return:
        """
        for record in records:
            if record.status != ValidityStatus.OK:
                logger = logging.getLogger(record.logger_name)
                limit_text = "under lower" if record.value < record.limit_value else "over upper"
                logger.warning(
                    'Variable "%s" out of bound: value %s%s is %s limit ( %s%s ) in file %s',
                    record.variable_name,
                    record.value,
                    " " + record.value_units if record.value_units else "",
                    limit_text,
                    record.limit_value,
                    " " + record.limit_units if record.limit_units else "",
                    record.source_file,
                )

    @classmethod
    def _update_problem_limit_definitions(cls, problem: om.Problem):
        """
        Updates limit definitions using variable declarations of provided OpenMDAO system.

        problem.setup() must have been run.

        :param problem:
        """

        variables = VariableList.from_problem(
            problem, get_promoted_names=False, promoted_only=False
        )
        for var in variables:
            system_path = var.name.split(".")
            system = problem.model
            for system_name in system_path[:-1]:
                system = getattr(system, system_name)
            var_name = system_path[-1]

            if hasattr(system, "_fastoad_limit_definitions"):

                limit_definitions = system._fastoad_limit_definitions

                if var_name in limit_definitions:
                    # Get units for already defined limits
                    limit_def = limit_definitions[var_name]
                    if limit_def.units is None and var.units is not None:
                        limit_def.units = var.units
                elif "lower" in var.metadata or "upper" in var.metadata:
                    # Get bounds if defined in add_output.
                    lower = var.metadata.get("lower")
                    # lower can be None if it is not found OR if it defined and set to None
                    if lower is None:
                        lower = -np.inf
                    upper = var.metadata.get("upper")
                    # upper can be None if it is not found OR if it defined and set to None
                    if upper is None:
                        upper = np.inf
                    units = var.metadata.get("units")
                    if lower > -np.inf or upper < np.inf:
                        limit_definitions[var_name] = _LimitDefinition(lower, upper, units)

    @staticmethod
    def _get_caller_filename():
        current_frame = inspect.currentframe()
        try:
            frame = inspect.getouterframes(current_frame)[2]
            filename = frame[1]
        finally:
            del current_frame
            del frame

        return filename


@dataclass
class _LimitDefinitions(dict):
    """
    Definition of validity check for one variable.
    """

    source_file: str
    logger_name: str


@dataclass
class _LimitDefinition:
    """
    Definition of one validity domain.

    namedtuple would not do, as we need the possibility to set values (especially
    units) after instantiation.
    """

    lower: float
    upper: float
    units: str = None
